from functools import partial
from typing import Callable

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.model_selection import StratifiedGroupKFold
from sklearn.base import is_classifier
from sklearn.metrics import check_scoring
from sklearn.utils import check_random_state
from sklearn.model_selection import check_cv
from skopt import BayesSearchCV
from optuna import logging
from optuna import samplers
from optuna import study as study_module
from optuna.integration.sklearn import OptunaSearchCV, _logger, _num_samples, _safe_indexing, _Objective, _check_fit_params, _convert_old_distribution_to_new_distribution
from optuna.trial import Trial

from luntaiDs.ModelingTools.utils.support import make_present_col_selector


def convert_subset_clf(base_estimator: BaseEstimator, features_subset: list) -> Pipeline:
    c = ColumnTransformer(
        transformers = [('subset', 'passthrough', make_present_col_selector(features_subset))], # pass all features show up in features_subset, no transformation is made, so 'passthrough'
        remainder = 'drop' # drop all remaining columns not presented in features_subset
    )
    pipe = Pipeline([
        ('transform', c),
        ('model', base_estimator)
    ])
    return pipe

class GroupStratifiedBayesSearchCV(BayesSearchCV):
    def __init__(self, estimator, group_col: str, search_spaces, n_cv: int = 3, optimizer_kwargs=None, n_iter=50,
                 scoring=None,
                 fit_params=None, n_jobs=1, n_points=1, refit=True, verbose=0,
                 pre_dispatch='2*n_jobs', random_state=None, error_score='raise', return_train_score=False):
        self.group_col = group_col
        self.n_cv = n_cv
        super().__init__(estimator=estimator, search_spaces=search_spaces, optimizer_kwargs=optimizer_kwargs,
                         n_iter=n_iter, scoring=scoring, fit_params=fit_params, n_jobs=n_jobs, n_points=n_points,
                         refit=refit, cv=StratifiedGroupKFold(n_splits=n_cv, random_state=random_state),
                         verbose=verbose, pre_dispatch=pre_dispatch,
                         random_state=random_state, error_score=error_score, return_train_score=return_train_score)

    def fit(self, X, y=None, *, groups=None, callback=None, **fit_params):
        if self.group_col not in X.columns:
            raise KeyError(f"{self.group_col} should be in training X")

        X_ = X.drop(columns=self.group_col)
        kfold_groups = X[self.group_col]
        return super().fit(X_, y, groups=kfold_groups, callback=callback, **fit_params)

    def score(self, X, y=None):
        return super().score(X.drop(columns=self.group_col), y)

    def score_samples(self, X):
        return super().score_samples(X.drop(columns=self.group_col))

    def predict(self, X):
        return super().predict(X.drop(columns=self.group_col))

    def predict_proba(self, X):
        return super().predict_proba(X.drop(columns=self.group_col))

    def predict_log_proba(self, X):
        return super().predict_log_proba(X.drop(columns=self.group_col))

    def decision_function(self, X):
        return super().decision_function(X.drop(columns=self.group_col))

    def transform(self, X):
        return super().transform(X.drop(columns=self.group_col))

class GroupStratifiedOptunaSearchCV(OptunaSearchCV):
    def __init__(self, estimator, group_col: str, param_distributions, n_cv: int = 3, enable_pruning=False, error_score=np.nan,
                 max_iter=1000, n_jobs=1, n_trials=10, random_state=None, refit=True, return_train_score=False,
                 scoring=None, study=None, subsample=1.0, timeout=None, verbose=0, callbacks=None):
        if not isinstance(param_distributions, dict):
            raise TypeError("param_distributions must be a dictionary.")
        self.group_col = group_col
        self.n_cv = n_cv

        self.cv = StratifiedGroupKFold(n_splits=n_cv, random_state=random_state)
        self.enable_pruning = enable_pruning
        self.error_score = error_score
        self.estimator = estimator
        self.max_iter = max_iter
        self.n_trials = n_trials
        self.n_jobs = n_jobs
        self.param_distributions = param_distributions
        self.random_state = random_state
        self.refit = refit
        self.return_train_score = return_train_score
        self.scoring = scoring
        self.study = study
        self.subsample = subsample
        self.timeout = timeout
        self.verbose = verbose
        self.callbacks = callbacks


    def fit(self, X, y=None, groups=None, **fit_params):
        if self.group_col not in X.columns:
            raise KeyError(f"{self.group_col} should be in training X")
        X_ = X.drop(columns=self.group_col)
        kfold_groups = X[self.group_col]

        # TODO: moved from __init__ to here, check if there are negative implications
        self.param_distributions = {
            key: _convert_old_distribution_to_new_distribution(dist)
            for key, dist in self.param_distributions.items()
        }
        return super().fit(X_, y, groups=kfold_groups, **fit_params)

    def score(self, X, y=None):
        return super().score(X.drop(columns=self.group_col), y)

    def predict(self, X):
        return super().predict(X.drop(columns=self.group_col))

    def predict_proba(self, X):
        return super().predict_proba(X.drop(columns=self.group_col))

    def predict_log_proba(self, X):
        return super().predict_log_proba(X.drop(columns=self.group_col))

    def decision_function(self, X):
        return super().decision_function(X.drop(columns=self.group_col))

    def transform(self, X):
        return super().transform(X.drop(columns=self.group_col))


import json
import logging
from collections import OrderedDict
import os


class EnhancedDict(OrderedDict):
    # customized dictionary

    def __init__(self, **kws):
        super().__init__(**kws)

    @ classmethod
    def fromDictSafe(cls, df):
        # convert from python original dictionary to EnhancedDict
        if isinstance(df, EnhancedDict):
            return df

        r = EnhancedDict()
        for k, v in df.items():
            if isinstance(v, dict):
                r[k] = EnhancedDict.fromDictSafe(v)
            else:
                r[k] = v
        return r

    def subset(self, keys):
        # extract specified key-value pairs
        r = EnhancedDict()
        for k, v in self.items():
            if k in keys:
                r[k] = v
        return r

    # override update method
    def update(self, E = None, append = True):
        if append == True:
            super().update(E)
        else:
            # if append = False, only update keys that are in self
            # for keys that are not in self, donnot add them
            for k, v in E.items():
                if k in self:
                    self[k] = v

    def merge(self, E = None, override = True):
        if override == True:
            super().update(E)
        else:
            # if override = False, only add keys that are not in self
            # for keys that are in self, keep the value unchanged
            for k, v in E.items():
                if k not in self:
                    self[k] = v

    def getKeyByValue(self, value, default=""):
        for k, v in self.items():
            if v == value:
                return k
        return default

    def checkValueType(self, vType):
        for k, v in self.items():
            if not isinstance(v, vType):
                raise TypeError("Please Ensure the format for %s is %s descendants!" % (k, vType))

    def applyFunc(self, keyFunc = lambda x:x, valueFunc = lambda x:x, inplace = False):
        '''
        apply uni-variable functions to key and value of the Enhanced dictionary repectively
        @param keyFunc: the univariate function to apply on the key (key transformation)
        @param valueFunc: the univariate function to apply on the value (value transformation)
        @param inplace: when inplace = True, no return and apply on itself, otherwise will return a transformed dict
        @return: only return when inplace = False
        '''
        r = EnhancedDict()
        for k, v in self.items():
            try:
                converted_k = keyFunc(k)
            except:
                logging.critical("Error when applying KeyFunc: %s on key %s; Ignore instead" % (keyFunc, k))
                converted_k = k
            try:
                converted_v = valueFunc(v)
            except:
                logging.critical("Error when applying ValueFunc: %s on key %s; Ignore instead" % (valueFunc, k))
                converted_v = v

            r[converted_k] = converted_v

        if inplace:
            self.clear()
            self.__init__(**r)
            return
        return r

    def println(self):
        for k, v in self.items():
            print(k, " => ", v)

class JsonDict(dict):
    # customized dictionary

    def __init__(self, **kws):
        super().__init__(**kws)

    @ classmethod
    def fromDictSafe(cls, df):
        # convert from python original dictionary to EnhancedDict
        if isinstance(df, JsonDict):
            return df

        if isinstance(df, str):
            df = json.loads(df)

        r = JsonDict()
        for k, v in df.items():
            if isinstance(v, dict):
                r[k] = JsonDict.fromDictSafe(v)
            else:
                r[k] = v
        return r

    def getDict(self, key):
        v = self.get(key)
        return self.fromDictSafe(v)


class Mappers(EnhancedDict):
    def __init__(self, **kws):
        super().__init__()
        for k, v in kws.items():
            self.bindMapper(k, v)

    def bindMapper(self, key, mapper):
        if callable(mapper):
            super().__setitem__(key, mapper)
        else:
            raise TypeError("mapper must be a function or a callable object!")

    def getMapper(self, key):
        return self.get(key)

    __setitem__ = bindMapper  # ensure not adding non-callable values

    __getitem__ = getMapper

    def applyMappers(self, df, defaultMapper=lambda x: x):
        '''
        @param df: input dictionary, e.g. input_dict = {factor1 : 100, factor2 : 200, factor3: 300}
        @param defaultMapper:
        @return:
        # will apply the mappers on each key-value pair in the input dictionary data
        Example:
        input_dict = {factor1 : 100, factor2 : 200, factor3: 300}
        self (mappers) = {factor1 : func1, factor2: func2}
        assume defaultMapper = funcD
        then:
        self.applyMappers(input_dict) -> {factor1 : func1(100), factor2: func2(200), factor3: funcD(300)}

        # Note that the number of key-value pairs in the output will be same as the input dictiondary!
        # use for score mapper!!!!!!!!!!!!!
        '''
        r = EnhancedDict()
        for k, v in df.items():
            mapper = self.getMapper(k)
            if not mapper:
                mapper = defaultMapper
            try:
                mapped_v = mapper(v)
            except:
                logging.critical("Error when apply map %s on key %s; Will keep value before mapping!" % (v, k))
                mapped_v = v
            r[k] = mapped_v
        return r

    def walkMappersOnDict(self, input_dict):
        '''
        @param input_dict: input dictionary, e.g. input_dict = {a : 4, b : 5, c : 6, d : 7}
        @return:
        # will apply all the mappers on one single dictionary data
        # assumes each mapper takes the whole dictionary data df as input
        Example:
        input_dict = {a : 4, b : 5, c : 6, d : 7}
        self (mappers) = {factor1 : func1, factor2: func2}
        then:
        self.walkMappersOnDict(input_dict) -> {factor1 : func1(input_dict), factor2: func2(input_dict)}

        # Note that the number of key-value pairs in the output will be same as the mapper itself!
        # use for factor mapper!!!!!!!!!!!!!
        '''
        factors = self.applyFunc(valueFunc=lambda mapper: mapper(input_dict))
        return factors
    

class SingleObjDataManager:
    def __init__(self, root_dir: str):
        self.ROOT_DIR = root_dir

    def get_path(self, *relative_path) -> str:
        return os.path.join(self.ROOT_DIR, *relative_path)

    def is_exist(self, *relative_path) -> bool:
        return os.path.exists(self.get_path(*relative_path))


from __future__ import annotations
import os
import shutil
from datetime import date
import time
from distutils.dir_util import copy_tree
from pathlib import Path
from typing import List
import logging
import pandas as pd
import pyspark
from tqdm.auto import tqdm
import ibis
from ibis.expr.schema import Schema
from ibis import schema
from luntaiDs.CommonTools.SnapStructure.tools import get_file_list_pattern, match_date_from_str
from luntaiDs.CommonTools.dbapi import baseDbInf, dbIO
from luntaiDs.CommonTools.dtyper import DSchema
from luntaiDs.CommonTools.sparker import SparkConnector
from luntaiDs.CommonTools.utils import str2dt
from luntaiDs.CommonTools.obj_storage import ObjStorage

    
class SnapshotDataManagerBase:
    @classmethod
    def setup(cls, spark_connector: SparkConnector = None, default_engine:str = 'pandas'):
        if spark_connector:
            cls.sc = spark_connector
        cls.default_engine = default_engine

    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema , e.g., RAW, PROCESSED
        :param table:  the virtual table under each schema, e.g., CLNT_GENERAL, BDA_GENERAL
        """
        self.schema = schema
        self.table = table

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(schema={self.schema}; table={self.table})"
    
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        raise NotImplementedError("")
    
    def init_table(self, *args, **kws):
        """initialize schema/table

        :return:
        """
        raise NotImplementedError("")
    
    def get_schema(self) -> Schema:
        """Check dtypes of the given schema/dataset

        :return:
        """

        if self.default_engine == 'pandas':
            df_random = self.read_random_snap()
            schema_ = []
            for n, t in zip(df_random.columns, df_random.dtypes):
                # https://numpy.org/doc/stable/reference/generated/numpy.dtype.kind.html
                if t.kind == 'O':
                    schema_.append((n, ibis.expr.datatypes.String()))
                else:
                    schema_.append((n, ibis.dtype(t)))
            return schema(schema_)
        if self.default_engine == 'spark':
            df_random = self.load_random_snap()
            return Schema.from_pyarrow(df_random.dtypes)
        else:
            raise ValueError("engine can only be spark or pandas")
    
    def pre_save_check(self, snap_dt: date, overwrite: bool = False) -> bool:
        """pre-check before inserting records, will delete existing records at snap_dt if overwrite is set to True

        :param snap_dt:
        :param overwrite: if False, will not insert record if exists; will delete otherwise
        :return: True = clear to go, False = exit in downstream
        """
        if overwrite:
            self.delete(snap_dt = snap_dt)
            return True
        else:
            num_records = self.count(snap_dt)
            if num_records > 0:
                logging.warning(f"Presave check failed, {num_records} existing records found for {self.schema}.{self.table}@{snap_dt} while overwrite is set to False")
                return False
            return True
    
    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError("")
    
    def save(self, df: pd.DataFrame, snap_dt: date, overwrite: bool = False, **kws):
        """save pandas dataframe to the system

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :param overwrite snap_dt: whether to overwrite the existing record if any
        :raises NotImplementedError: _description_
        """
        if not self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite):
            return
        
        logging.info(f"Saving dataframe, shape {df.shape}, into {self.schema}.{self.table}")
        self._save(df = df, snap_dt = snap_dt, **kws)
        logging.info(f"Successfully saved {len(df)} records to {self.schema}.{self.table}@{snap_dt}")

    def _save_partitions(self, df: pyspark.sql.DataFrame, snap_dt: date, **kws):
        """the pure logic to save pyspark dataframe to the system, without handling existing record problem

        :param pyspark.sql.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        raise NotImplementedError("")

    def save_partitions(self, df: pyspark.sql.DataFrame, snap_dt: date, overwrite: bool = False, **kws):
        """save pyspark dataframe to the system

        :param pyspark.sql.DataFrame df: _description_
        :param date snap_dt: _description_
        :param overwrite snap_dt: whether to overwrite the existing record if any
        :raises NotImplementedError: _description_
        """
        if not self.pre_save_check(snap_dt = snap_dt, overwrite = overwrite):
            return
        
        self._save_partitions(df = df, snap_dt = snap_dt, **kws)
        logging.info(f"Successfully saved {df.count()} records to {self.schema}.{self.table}@{snap_dt}")

    def etl_from_db(self, db_conf: baseDbInf, sql: str, snap_dt: date, engine: str = 'pandas'):
        """do ETL work, extracting data from database and save it to local dir

        :param db_conf: the database configuration object, specify which db and connection param
        :param sql: the sql to do the ETL
        :param snap_dt: the snap date to do the ETL, for storage purpose
        :param engine: the engine to query_extract {pandas, spark}
        :return:
        """
        if engine == "pandas":
            db_conn = dbIO(db_conf)
            df = db_conn.query_sql_df(sql)
            self.save(df, snap_dt)
        elif engine == "spark":
            # spark data query_extract
            if hasattr(self, "sc"):
                df = self.sc.query_db(db_conf, sql)
                self.save_partitions(df, snap_dt)
            else:
                raise ValueError("No spark session binded, please bind via .setup() method")
        else:
            raise ValueError("Engine can only be pandas or spark")
        
    def get_existing_snap_dts(self) -> List[date]:
        raise NotImplementedError("")
    
    def count(self, snap_dt: date) -> int:
        """count the sample size for given snap date

        :param date snap_dt: _description_
        :return int: # of rows of the dataframe
        """
        if snap_dt not in self.get_existing_snap_dts():
            return 0
        if self.default_engine == 'pandas':
            df = self.read(snap_dt=snap_dt)
            return len(df)
        if self.default_engine == 'spark':
            df = self.load(snap_dt=snap_dt)
            return df.count()
        else:
            raise ValueError("engine can only be spark or pandas")
    
    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        raise NotImplementedError("")
    
    def reads(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        raise NotImplementedError("")

    def read_random_snap(self) -> pd.DataFrame:
        """Select a random snap from the schema.table

        :return:
        """
        import random

        existing_snaps = self.get_existing_snap_dts()
        random_snap = random.choice(existing_snaps)
        return self.read(random_snap)
    
    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        raise NotImplementedError("")

    def load_random_snap(self) -> pyspark.sql.DataFrame:
        """Select a random snap from the schema.table

        :return:
        """
        import random

        existing_snaps = self.get_existing_snap_dts()
        random_snap = random.choice(existing_snaps)
        return self.load(random_snap)
    
    def loads(self, snap_dts: List[date], **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        raise NotImplementedError("")
    
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        raise NotImplementedError("")

    def drop(self):
        """drop the whole table

        :return:
        """
        raise NotImplementedError("")

    def migrate(self, dst_schema: str, dst_table: str):
        """move from the existing schema.table to new one, the existing one will be deleted

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = self.duplicate(dst_schema, dst_table)
        self.drop()
        return new

    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerBase:
        """duplicate the existing schema.table to new one, the existing one will be kept

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        raise NotImplementedError("")

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date on disk

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        raise NotImplementedError("")
    
    def is_valid(self, snap_dt, rows_threshold: int = 100) -> bool:
        rows = self.count(snap_dt = snap_dt)
        return rows > rows_threshold
    


class SnapshotDataManagerLocalFS(SnapshotDataManagerBase):
    """files are saved as parquet snapshots under each schema.table
        file naming convention: dir/tablename_YYYY-MM-DD.parquet
    """

    @classmethod
    def setup(cls, root_dir: str, spark_connector: SparkConnector = None, default_engine:str = 'pandas'):
        super(SnapshotDataManagerLocalFS, cls).setup(
            spark_connector = spark_connector,
            default_engine = default_engine
        )
        cls.ROOT_DIR = root_dir
        cls.default_engine = default_engine
        
    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema
        :param table:  the virtual table under each schema
        """
        super().__init__(schema = schema, table = table)
        dir = os.path.join(self.ROOT_DIR, schema, table)
        self.dir = dir  # the root path of the table under each schema
        self.init_table()
        
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        return os.path.exists(self.dir)

    def init_table(self):
        # https://note.nkmk.me/en/python-os-mkdir-makedirs/
        os.makedirs(self.dir, exist_ok = True)

    def get_filename(self, snap_dt: date) -> str:
        return f"{self.table}_{snap_dt}.parquet"

    def get_default_file_path(self, snap_dt: date) -> str:
        return os.path.join(self.dir, self.get_filename(snap_dt))
    

    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        filepath = self.get_default_file_path(snap_dt)
        df.to_parquet(filepath, **kws)
        logging.info(f"Successfully saved file to {filepath}")

    def _save_partitions(self, df: pyspark.sql.DataFrame, snap_dt: date, **kws):
        """the pure logic to save pyspark dataframe to the system, without handling existing record problem

        :param pyspark.sql.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        filepath = self.get_default_file_path(snap_dt)
        self.sc.save_parquet(df, filepath, **kws)
        logging.info(f"Successfully saved file to {filepath}")

    def get_existing_snap_dts(self) -> List[date]:
        existing_files = get_file_list_pattern(
            self.dir,
            pattern = f"*{self.table}*",
            include_dir = False  # don't include directory path
        )

        existing_snaps = list(set(str2dt(match_date_from_str(f)) for f in existing_files))
        existing_snaps.sort()
        return existing_snaps

    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_default_file_path(snap_dt)
        return pd.read_parquet(filepath, **kws)

    def reads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        # do the checks to ensure successful concatenation
        li =  []
        columns = pd.Index([])
        for snap_dt in tqdm(snap_dts):
            df_partition = self.read(snap_dt=snap_dt, **kws)
            if partition_key is not None and partition_key not in df_partition.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found in {snap_dt}")
            if len(columns) > 0 and len(columns.union(df_partition.columns)) != len(columns):
                # if new column set has a different length than the previous one, means the columns are not same across all the table partitions
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent for snapshot {snap_dt}")
            li.append(df_partition)

        return pd.concat(li, axis = 0, ignore_index = True)


    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_default_file_path(snap_dt)
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(filepath, **kws)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")

    def loads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        file_paths = [self.get_default_file_path(snap_dt) for snap_dt in snap_dts]
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(*file_paths, **kws) # spark will automatically stack all tables together
            # do the partition check
            if partition_key is not None and partition_key not in df.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found")

            # pick up random snapshot data to check the columns are consistent
            one_columns = self.load_random_snap().columns
            if len(one_columns) != len(pd.Index(one_columns).union(df.columns)):
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent in one of the snapshot data")

            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")

    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        filepath = self.get_default_file_path(snap_dt)
        if os.path.exists(filepath):
            if os.path.isdir(filepath):
                # parquet partitions
                shutil.rmtree(filepath, ignore_errors = True)
            else:
                os.remove(filepath)

    def drop(self):
        """drop the whole table

        :return:
        """
        shutil.rmtree(self.dir)

    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerLocalFS:
        new = SnapshotDataManagerLocalFS(schema = dst_schema, table = dst_table)
        copy_tree(self.dir, new.dir)
        return new

    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date file (pandas) or folder (pyspark partitions)

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        f = self.get_default_file_path(snap_dt)
        if os.path.isfile(f):
            size_bytes = os.path.getsize(f)
        else:
            size_bytes = sum(p.stat().st_size for p in Path(f).rglob('*'))

        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size

    def is_valid(self, snap_dt, size_threshold: float = 0.5) -> bool:
        """Check if the file is valid
        
        """
        f = self.get_default_file_path(snap_dt)
        if os.path.isdir(f):
            # if it is a folder (partition files), it should has a .__SUCCESS.crc file generated by pyspark
            return "._SUCCESS.crc" in os.listdir(f)
        else:
            # if a single file, the size must be exceed the valid threshold
            return self.disk_space(snap_dt, unit = 'MB') > size_threshold
        
        
class SnapshotDataManagerObjStorage(SnapshotDataManagerBase):
    """files are saved as object storage under each schema.table
    """

    @classmethod
    def setup(cls, bucket: str,  root_dir: str, obja: ObjStorage, spark_connector: SparkConnector = None, default_engine:str = 'pandas'):
        super(SnapshotDataManagerObjStorage, cls).setup(
            spark_connector = spark_connector,
            default_engine = default_engine
        )
        cls.bucket = bucket
        cls.ROOT_DIR = root_dir if not root_dir.startswith("/") else root_dir[1:]
        cls.obja = obja
        cls.obja.enter_bucket(bucket_name=bucket)
        
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        return True
    
    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema
        :param table:  the virtual table under each schema
        """
        super().__init__(schema = schema, table = table)
        dir = os.path.join(self.ROOT_DIR, schema, table)
        self.dir = dir  # the root path of the table under each schema

    def get_filename(self, snap_dt: date) -> str:
        return f"{self.table}_{snap_dt}.parquet"

    def get_default_file_path(self, snap_dt: date) -> str:
        return os.path.join(self.dir, self.get_filename(snap_dt))
    
    def get_file_path_with_obj_storage_prefix(self, snap_dt: date) -> str:
        return f"{self.get_uri_protocol}://{self.bucket}/{self.get_default_file_path(snap_dt=snap_dt)}"

    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        self.obja.save_parquet(
            df = df,
            remote_path = self.get_default_file_path(snap_dt=snap_dt),
            **kws
        )
    
    def get_existing_snap_dts(self) -> List[date]:
        df = self.obja.list_objs(remote_path = self.dir)
        if df is None:
            return []
        existing_snaps = list(set(map(
            str2dt,
            (
                df
                .loc[
                    (df['Key'].str.startswith(pat = f"{self.dir}/"))
                    & (df['Key'].str.endswith(pat = ".parquet")), 
                    'Key']
                .str.replace(self.dir, "")
                .apply(match_date_from_str)
                .tolist()
            )
        )))
        existing_snaps.sort()
        return existing_snaps
    
    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_default_file_path(snap_dt=snap_dt)
        return self.obja.read_parquet(remote_path = filepath, **kws)
    

    def reads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        # do the checks to ensure successful concatenation
        li =  []
        columns = pd.Index([])
        for snap_dt in tqdm(snap_dts):
            df_partition = self.read(snap_dt=snap_dt, **kws)
            if partition_key is not None and partition_key not in df_partition.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found in {snap_dt}")
            if len(columns) > 0 and len(columns.union(df_partition.columns)) != len(columns):
                # if new column set has a different length than the previous one, means the columns are not same across all the table partitions
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent for snapshot {snap_dt}")
            li.append(df_partition)

        return pd.concat(li, axis = 0, ignore_index = True)
    
    def load(self, snap_dt: date, **kws) -> pyspark.sql.DataFrame:
        """Read as spark dataframe (one snapshot date) data, and can also access from sc temporary view

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_file_path_with_obj_storage_prefix(snap_dt)
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(filepath, **kws)
            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")
    
    def loads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pyspark.sql.DataFrame:
        """reads as pyspark dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        file_paths = [self.get_file_path_with_obj_storage_prefix(snap_dt) for snap_dt in snap_dts]
        if hasattr(self, "sc"):
            df = self.sc.read_parquet(*file_paths, **kws) # spark will automatically stack all tables together
            # do the partition check
            if partition_key is not None and partition_key not in df.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found")

            # pick up random snapshot data to check the columns are consistent
            one_columns = self.load_random_snap().columns
            if len(one_columns) != len(pd.Index(one_columns).union(df.columns)):
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent in one of the snapshot data")

            df.createOrReplaceTempView(f"{self.table}")
            return df
        else:
            ValueError("No Spark Connector Specified, please call .setup() to bind a spark connector")
    
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe, could be partitions

        :param snap_dt: which snap date to delete
        :return:
        """
        filepath = self.get_default_file_path(snap_dt=snap_dt)
        if self.obja.list_objs(filepath) is None or len(self.obja.list_objs(filepath)) == 1:
            # it is not folder
            self.obja.delete_obj(remote_path = filepath)
        else:
            # it is a folder, i.e., partitions
            self.obja.delete_folder(remote_folder_path = filepath)

    def drop(self):
        """drop the whole table

        :return:
        """
        self.obja.delete_folder(remote_folder_path = self.dir)

    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerObjStorage:
        """duplicate the existing schema.table to new one, the existing one will be kept

        :param dst_schema: destination schema
        :param dst_table: destination table
        :return:
        """
        new = SnapshotDataManagerObjStorage(schema = dst_schema, table = dst_table)
        for snap_dt in tqdm(self.get_existing_snap_dts()):
            from_filename = self.get_default_file_path(snap_dt=snap_dt)
            to_filename = new.get_default_file_path(snap_dt=snap_dt)
            self.obja.copy_obj(
                from_path = from_filename,
                to_path = to_filename
            )
        return new
        
    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date on disk

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        #key = self.obja.get_obj(remote_path=self.get_default_file_path(snap_dt=snap_dt))
        #size_bytes = key['ContentLength'] # in kb
        size_bytes = self.obja.list_objs(remote_path=self.get_default_file_path(snap_dt=snap_dt))['Size'].sum()
        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size
    
    def is_valid(self, snap_dt, size_threshold: float = 0.5) -> bool:
        """Check if the file is valid
        
        """
        return self.disk_space(snap_dt, unit = 'MB') > size_threshold
    
    
        
class SnapshotDataManagerWarehouseMixin(SnapshotDataManagerBase):
    """Mixin class for warehouse implementation (partial)
    """
    def __init__(self, schema:str, table:str, snap_dt_key: str):
        """database management interface

        :param schema: schema
        :param table:  table under each schema
        :param snap_dt_key: snap date column name for all tables
        """
        super().__init__(schema = schema, table = table)
        self.snap_dt_key = snap_dt_key
        
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        return self.is_exist(schema=self.schema, table=self.table)
    
    def init_table(self, col_schemas: DSchema, overwrite:bool = False, **settings):
        """initialize/create table in the underlying data warehouse system

        :param DSchema col_schemas: data column schema
        :param bool overwrite: whether to drop table if exists, defaults to False
        """

        if self.exist():
            if overwrite:
                self.drop()
            else:
                logging.warning(f"{self.schema}.{self.table} already exists, will do nothing." 
                                "set overwrite to True if you wish to reset table")
                return
        
        # create schema
        self.create_schema(schema = self.schema)
        # create table
        self.create_table(
            schema = self.schema,
            table = self.table,
            col_schemas = col_schemas,
            **settings
        )
        
    def get_schema(self) -> Schema:
        """Check dtypes of the given schema/dataset

        :return:
        """
        return self.get_dtypes(
            schema = self.schema,
            table = self.table
        )
        
    def count(self, snap_dt: date) -> int:
        if not self.exist():
            return 0
        table = self.get_table(
            schema = self.schema,
            table = self.table
        )
        return table.count(
            where = (table[self.snap_dt_key] == snap_dt)
        ).to_pandas() # convert to a scalar number
        
    def get_existing_snap_dts(self) -> List[date]:
        if not self.exist():
            return []
        existing_snaps = (
            self.get_table(
                schema = self.schema,
                table = self.table
            ).select(self.snap_dt_key)
            .distinct()
            .order_by(self.snap_dt_key)
            .to_pandas()
        )
        if len(existing_snaps) == 0:
            return []
        return list(
            str2dt(dt.date()) 
            for dt in pd.to_datetime(existing_snaps[self.snap_dt_key]).dt.to_pydatetime()
        )
            
    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        table = self.get_table(
            schema = self.schema,
            table = self.table
        )
        df = table.filter(table[self.snap_dt_key] == snap_dt)
        if 'columns' in kws:
            df = df.select(*kws['columns'])
        return df.to_pandas()
    
    def reads(self, snap_dts: List[date], **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        table = self.get_table(
            schema = self.schema,
            table = self.table
        )
        df = table.filter(table[self.snap_dt_key].isin(snap_dts))
        if 'columns' in kws:
            df = df.select(*kws['columns'])
        return df.to_pandas()

    def drop(self):
        """drop the whole table

        :return:
        """
        self.delete_table(
            schema = self.schema,
            table = self.table,
        )


import io
import os

from botocore.exceptions import ClientError
import logging
import pandas as pd
import boto3
from luntaiDs.CommonTools.obj_storage import ObjStorage, remove_begin_slash, add_tail_slash

class S3Accessor(ObjStorage):
    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                 aws_session_token=None, region_name=None,
                 botocore_session=None, profile_name=None,
                 service_name:str = 's3', api_version=None,
                 use_ssl=None, verify=None, endpoint_url=None,
                 config=None
                 ):
        self.region_name = region_name
        try:
            self.session = boto3.Session(
                aws_access_key_id = aws_access_key_id,
                aws_secret_access_key = aws_secret_access_key,
                aws_session_token = aws_session_token,
                region_name = region_name,
                botocore_session = botocore_session,
                profile_name = profile_name
            )

            self.client = self.session.client(
                service_name, api_version = api_version, use_ssl = use_ssl,
                verify = verify, endpoint_url = endpoint_url,
                config = config
            )
            self.resource = self.session.resource(
                service_name, api_version = api_version, use_ssl = use_ssl,
                verify = verify, endpoint_url = endpoint_url,
                config = config
            )

        except Exception as e:
            logging.error(f"Boto3 failed to create client to access {service_name} \n{e}")
        else:
            logging.info("Boto3 session/client/resource created successfully.")
            
    def get_uri_protocol(self) -> str:
        return "s3a"

    def list_buckets(self) -> pd.DataFrame:
        return pd.DataFrame.from_records(self.client.list_buckets()['Buckets'])

    def create_bucket(self, bucket_name: str) -> int:
        """Create a new bucket

        :param bucket_name: name of the bucket
        :return:
        """
        if self.region_name:
            region_config = {'LocationConstraint': self.region_name}
        else:
            region_config = None
        self.client.create_bucket(
            Bucket = bucket_name,
            CreateBucketConfiguration = region_config
        )
        return 1

    def get_bucket(self, bucket_name: str = None):
        return self.resource.Bucket(bucket_name)

    def get_buckets(self):
        return self.resource.buckets

    def get_bucket_iter(self):
        for bucket in self.resource.buckets:
            yield bucket

    def enter_bucket(self, bucket_name: str = None) -> int:
        """move to a given bucket

        :param bucket_name: name of the bucket you want to switch to
        :return:
        """
        # check if exists
        if self.resource.Bucket(bucket_name) in self.resource.buckets.all():
            self.BUCKET = bucket_name
            return 1
        else:
            logging.error(f"Bucket {bucket_name} does not exists, please call .list_buckets() to see all available buckets")
            return 0

    def if_not_exist_return_default_bucket(self, bucket_name = None):
        return self.BUCKET if bucket_name is None else bucket_name


    def get_current_bucket(self):
        """get your current bucket object

        :return:
        """
        if hasattr(self, 'BUCKET'):
            return self.resource.Bucket(self.BUCKET)
        else:
            raise ValueError("Has not enter into any bucket yet, please call .enter_bucket(bucket_name) to enter a specific bucket")

    def delete_bucket(self, bucket_name: str, interactive: bool = True) -> int:
        """delete the bucket

        :param bucket_name: name of the bucket to delete
        :param interactive: whether you would like an interactive warning message
        :return:
        """
        logging.warning(f"You are about to delete bucket: {bucket_name} and everything under it, please confirm")
        if interactive:
            y_n = input(f"delete everything? (y/n)")
            if y_n in ['y', 'Y']:
                self.resource.Bucket(bucket_name).objects.all().delete()
                self.resource.Bucket(bucket_name).delete()
                return 1
            else:
                logging.warning(f"Aborted by user, please try again")
                return 0

        self.resource.Bucket(bucket_name).delete()
        return 1


    ''' below requires lock a bucket '''

    def list_objs(self, remote_path: str, bucket_name: str = None) -> pd.DataFrame:
        """list objects under the specific directory

        :param remote_path: remote path excluding bucket
        :param bucket_name: name of the bucket, if not given, will use the default bucket
        :return: pandas dataframe showing files inside
        """
        pd.set_option('display.max_colwidth', None)

        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        remote_path = remove_begin_slash(remote_path)

        try:
            res = self.client.list_objects(Bucket = bucket_name, Prefix = remote_path)
        except ClientError as e:
            logging.error(f'Error occurred when listing objects/files under path: {bucket_name}/{remote_path}: \n{e}')
        else:
            if 'Contents' not in res.keys():
                print(f"No objects found at {bucket_name}/{remote_path}")
            else:
                df = pd.DataFrame.from_records(res['Contents'])

                logging.info(f"There are {len(res['Contents'])} files found in s3a://{bucket_name}/{remote_path}")
                return df[['Key', 'LastModified', 'Size']]

    def get_obj(self, remote_path:str, bucket_name: str = None) -> dict:
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        obj = self.client.get_object(Bucket = bucket_name, Key= remove_begin_slash(remote_path))
        return obj

    def create_obj(self, remote_path: str, bucket_name: str = None) -> int:
        """Create an object

        :param remote_path: if remote path ends with /, it will be folder, otherwise will be a blank file
        :param bucket_name:
        :return: 1 if success otherwise the info dictionary
        """
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        response  = self.client.put_object(
            Bucket = bucket_name,
            Key = remote_path
        )

        if response.get('ResponseMetadata', {}).get('HTTPStatusCode', {}) == 200:
            logging.info(f"Object Created in Bucket: {bucket_name}, path: {remote_path} ")
            return 1
        else:
            return response

    def delete_obj(self, remote_path: str, bucket_name: str = None) -> int:
        """Delete an object/file

        :param remote_path: the filepath to delete (excluding bucket path)
        :param bucket_name: in which bucket
        :return:
        """
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)
        self.resource.Object(bucket_name, remove_begin_slash(remote_path)).delete()
        return 1

    def create_folder(self, remote_folder_path: str, bucket_name: str = None) -> int:
        """Create a folder object on S3

        :param remote_folder_path: the folder path (excluding bucket path)
        :param bucket_name: under which bucket
        :return:
        """
        remote_folder_path = add_tail_slash(remote_folder_path)
        response = self.create_obj(remote_folder_path, bucket_name)
        return response

    def delete_folder(self, remote_folder_path: str, bucket_name: str = None) -> int:
        """Delete the folder and every file underneath it

        :param remote_folder_path: the folder path to delete (excluding bucket path)
        :param bucket_name:
        :return:
        """
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        remote_folder_path = add_tail_slash(remote_folder_path)
        self.resource.Bucket(bucket_name).objects.filter(Prefix = remote_folder_path).delete()
        return 1

    def copy_obj(self, from_bucket_name: str = None, from_path: str = None, to_bucket_name: str = None, to_path: str = None):
        """Copy one file from one place to another (can copy to a different bucket)

        :param from_bucket_name: from which bucket you want to copy the file
        :param from_path: which file you want to copy (excluding bucket path)
        :param to_bucket_name: to which bucket you want to copy to
        :param to_path: the destination file path (excluding bucket path)
        :return:
        """
        from_bucket_name = self.if_not_exist_return_default_bucket(from_bucket_name)
        to_bucket_name = self.if_not_exist_return_default_bucket(to_bucket_name)

        copy_source = {
            'Bucket': from_bucket_name,
            'Key': remove_begin_slash(from_path)
        }
        self.client.copy(copy_source, to_bucket_name, remove_begin_slash(to_path))


    def upload_file(self, local_path: str, remote_path: str, bucket_name: str = None, multipart_threshold: int = 1000000000000):
        '''upload file from local machine to S3

        :param local_path: local file path that you would like to upload to S3, suggest to be absolute path
        :param remote_path: remote path on MinIO (excluding bucket path); could either start with / or not
        :param multipart_threshold: Setting an extremely large multipart threshold to effectively disable multipart uplaod.
                Note: Multiplart upload is not compatible with GCS, therefore disabling multipart upload for compatability with hybrid solution.
                set it to None is you want multipart saving
        :return:
        '''
        key = remove_begin_slash(remote_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        try:
            if multipart_threshold:
                config = boto3.s3.transfer.TransferConfig(multipart_threshold = multipart_threshold)
                self.client.upload_file(local_path, bucket_name, key, Config = config)
            else:
                self.client.upload_file(local_path, bucket_name, key)
        except Exception as e:
            logging.error(f'File upload FAILED from {local_path} to {key}:\n'+ str(e))
        else:
            logging.info(f'File uploaded successfully from {local_path} to {key}:')

    def download_file(self, remote_path: str, local_path: str, bucket_name: str = None):
        '''download file from S3 to local machine

        :param remote_path: remote path on S3 (excluding bucket); could either start with / or not
        :param local_path: local file path that you would like to save the file from MinIO, suggest to be absolute path
        :param bucket_name
        :return:
        '''
        key = remove_begin_slash(remote_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        try:
            self.client.download_file(bucket_name, key, local_path)
        except Exception as e:
            logging.error(f'File download FAILED from {key} to {local_path}:\n'+ str(e))
        else:
            logging.info(f'File download successfully from {key} to {local_path}:')


    def read_obj(self, remote_path: str, bucket_name: str = None) -> bytes:
        """read obj from S3 without downloading

        :param remote_path:  remote path to read on S3 (excluding bucket); could either start with / or not
        :param Bucket:
        :return: a buffer (io.StringIO() or io.BytesIO() which can pass to file reader
        """

        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        obj = self.client.get_object(Bucket = bucket_name, Key= remove_begin_slash(remote_path))
        return obj['Body'].read()

    def save_obj(self, body: bytes, remote_path: str, bucket_name: str = None) -> 1:
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        self.resource.Object(bucket_name, remove_begin_slash(remote_path)).put(Body = body)
        return 1

    def read_obj_to_iobuffer(self, remote_path: str, bucket_name: str = None) -> io.BytesIO:
        obj = self.read_obj(remote_path, bucket_name)
        buffer = io.BytesIO(obj)
        return buffer

    def save_iobuffer_to_obj(self, buffer: io.BytesIO, remote_path: str, bucket_name: str = None):
        self.save_obj(buffer.getvalue(), remote_path, bucket_name)

    def read_csv(self, remote_path: str, bucket_name: str = None, *args, **kws) -> pd.DataFrame:
        """Read CSV from S3 directly into Pandas

        :param remote_path: csv file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.read_csv(*args, **kws)
        :param kws: the keyword args passed to pd.read_csv(*args, **kws)
        :return: pandas dataframe
        """
        buffer = self.read_obj_to_iobuffer(remote_path, bucket_name)
        df = pd.read_csv(buffer, *args, **kws)
        return df

    def save_csv(self, df: pd.DataFrame, remote_path: str, bucket_name: str = None, *args, **kws):
        """Save pandas dataframe directly to csv on S3

        :param df: pandas dataframe
        :param remote_path: csv file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.to_csv(*args, **kws)
        :param kws: the keyword args passed to pd.to_csv(*args, **kws)
        :return:
        """
        buffer = io.BytesIO()
        df.to_csv(buffer, *args, **kws)

        self.save_iobuffer_to_obj(buffer, remote_path, bucket_name)

    def read_parquet(self, remote_path: str, bucket_name: str = None, *args, **kws) -> pd.DataFrame:
        """Read Parquet file from S3 directly into Pandas

        :param remote_path: parquet file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.read_parquet(*args, **kws)
        :param kws: the keyword args passed to pd.read_parquet(*args, **kws)
        :return:
        """
        buffer = self.read_obj_to_iobuffer(remote_path, bucket_name)
        df = pd.read_parquet(buffer, *args, **kws)
        return df

    def save_parquet(self, df: pd.DataFrame, remote_path: str, bucket_name: str = None, *args, **kws):
        """Save pandas dataframe directly to parquet file on S3

        :param df: pandas dataframe
        :param remote_path: parquet file path (excluding bucket path)
        :param bucket_name:
        :param args: the positional args passed to pd.to_parquet(*args, **kws)
        :param kws: the keyword args passed to pd.to_parquet(*args, **kws)
        :return:
        """
        buffer = io.BytesIO()
        df.to_parquet(buffer, *args, **kws)

        self.save_iobuffer_to_obj(buffer, remote_path, bucket_name)

    def upload_folder(self, local_folder_path: str, remote_root_path: str, bucket_name: str = None, multipart_threshold: int = 1000000000000):
        """Upload a local folder to S3

        :param local_folder_path: local folder path (full path)
        :param remote_root_path: root path where the folder will be created on S3 (excluding bucket path)
        :param bucket_name:
        :param multipart_threshold: an argument controlling uploading size per partition
        :return:
        """
        local_folder_path = os.path.abspath(local_folder_path)
        local_parent_dir_path = os.path.dirname(local_folder_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        for path, subdirs, files in os.walk(local_folder_path):

            dirname = path.replace(local_parent_dir_path, '')
            for file in files:
                local_filepath = os.path.join(path, file)
                remote_filepath = os.path.join(dirname, file)
                if os.path.sep == '\\':
                    remote_filepath = remote_filepath.replace("\\", "/")
                remote_filepath = add_tail_slash(remove_begin_slash(remote_root_path)) +  remove_begin_slash(remote_filepath)

                if multipart_threshold:
                    config = boto3.s3.transfer.TransferConfig(multipart_threshold = multipart_threshold)
                    self.client.upload_file(local_filepath, bucket_name, remote_filepath, Config = config)
                else:
                    self.client.upload_file(local_filepath, bucket_name, remote_filepath)

        logging.info(f"Successfully upload folder from {local_folder_path} to {remote_root_path}")


    def download_folder(self, remote_folder_path: str, local_root_path: str, bucket_name: str = None):
        """

        :param remote_folder_path: folder path on S3 (excluding bucket path)
        :param local_root_path: local root path where the folder will be downloaded to (full root path)
        :param bucket_name:
        :return:
        """
        remote_folder_path = remove_begin_slash(remote_folder_path)
        bucket_name = self.if_not_exist_return_default_bucket(bucket_name)

        for obj in self.resource.Bucket(bucket_name).objects.filter(Prefix = remote_folder_path):
            remote_filepath = obj.key
            local_filepath = os.path.join(local_root_path, os.path.relpath(remote_filepath, remote_folder_path))
            if not os.path.exists(os.path.dirname(local_filepath)):
                os.makedirs(os.path.dirname(local_filepath))
            if remote_filepath.endswith("/"):  # bypass void path
                continue

            self.client.download_file(bucket_name, remote_filepath, local_filepath)

        logging.info(f"Successfully download folder from {remote_folder_path} to {local_root_path}")



from __future__ import annotations
from collections import OrderedDict
import json
import shutil
import os
from typing import List, Dict
import pandas as pd
from datetime import datetime
from luntaiDs.CommonTools.accessor import loadJSON, toJSON
from luntaiDs.CommonTools.obj_storage import ObjStorage

class TimeInterval:
    def __init__(self, start: datetime, end: datetime):
        if start >= end:
            raise ValueError("Start must be ealier than end datetime")
        self.start = start
        self.end = end

    def __str__(self) -> str:
        return f"TimeInterval({self.start} - {self.end})"
    
    def __contains__(self, dt: datetime) -> bool:
        return self.start <= dt <= self.end

    def overlap(self, ti: TimeInterval) -> bool:
        return ti.end > self.start and self.end > ti.start
    
    def overlap_type(self, ti: TimeInterval) -> int:
        """judge which type is the overlapping

        there are five types
        0 - no overlapping
        1 - ti is left of self
        2 - ti is right of self
        3 - ti includes self
        4 - self includes ti
        5 - perfect overlap

        :param TimeInterval ti: _description_
        :return int: {0,1,2,3,4}
        """
        if not self.overlap(ti):
            return 0
        elif  ti.start == self.start and self.end == ti.end:
            return 5
        elif ti.start < self.start < ti.end < self.end:
            return 1
        elif self.start < ti.start < self.end < ti.end:
            return 2
        elif ti.start <= self.start < self.end <= ti.end:
            return 3
        elif self.start <= ti.start < ti.end <= self.end:
            return 4

    def to_js(self) -> Dict[str, str]:
        return {
            'start' : self.start.isoformat(), 
            'end' : self.end.isoformat()
        }
    
    @classmethod
    def from_js(cls, l: Dict[str, str]) -> TimeInterval:
        return TimeInterval(
            start = datetime.fromisoformat(l['start']),
            end = datetime.fromisoformat(l['end'])
        )


class _BaseModelTimeTable:
    """you can have different model (model_id) run in different time period as prod model
    support for multiple schedules
    """
    def load_tb(self) -> Dict[str, List[TimeInterval]]:
        """load time table

        :return Dict[str, List[TimeInterval]]: time table loaded
        """
        raise NotImplementedError("")
    
    def save_tb(self, tb: Dict[str, List[TimeInterval]]):
        """save timetable

        :param Dict[str, List[TimeInterval]] tb: time table
        """
        raise NotImplementedError("")

    @ classmethod
    def sort_and_merge(cls, timetable: Dict[str, List[TimeInterval]]) -> Dict[str, List[TimeInterval]]:
        """sort and merge timetable to avoid any overlapping conflict

        :param Dict[str, List[TimeInterval]] timetable: timetable to be sorted
        :return Dict[str, List[TimeInterval]]: sorted and merged timetable
        """
        r = {}
        for mi, tis in timetable.items():
            length = len(tis)
            if length == 1:
                r[mi] = tis
                continue

            sort_tis = sorted(tis, key=lambda x: x.start)
            s = []; i = 0
            while i < length:
                cur = sort_tis[i]
                if i + 1 < length:
                    nex = sort_tis[i + 1]
                    if cur.end == nex.start:
                        cur.end = nex.end
                        i += 1
                s.append(cur)
                i += 1

            r[mi] = s
        return r


    def register(self, model_id: str, start: datetime, end: datetime,  force: bool = False):
        """add a new time interval into the schedule

        :param str model_id: model id to be registered
        :param datetime start: start datetime of this model id
        :param datetime end: end datetime of this model id
        :param bool force: if True, will move endpoints of existing periods to avoid overlapping
        """
        new = TimeInterval(start=start, end=end)
        timetable = self.load_tb()
        for mi, tis in timetable.items():
            del_list = []
            for i, ti in enumerate(tis.copy()):
                if new.overlap(ti):
                    if force is False:
                        raise ValueError(f"Overlap with existing model {mi} serving time {ti}, set force to True or change endpoints")
                    else:
                        ti_end = ti.end
                        ti_start = ti.start
                        ovt = new.overlap_type(ti)
                        if ovt == 1:
                            # end of ti -> start of new
                            timetable[mi][i].end = start
                        elif ovt == 2:
                            # start of ti -> end of new
                            timetable[mi][i].start = end
                        elif ovt == 3:
                            # end of ti -> start of new 
                            # + create new interval [end of new, end of ti]
                            timetable[mi][i].end = start
                            timetable[mi].append(TimeInterval(start=end, end=ti_end))
                        elif ovt == 4 or ovt == 5:
                            # remove ti
                            del_list.append(i)
                            # del timetable[mi][i]  # will cause index problem when deleting
            
            # delete task
            cleaned = [timetable[mi][i] for i in range(len(timetable[mi])) if i not in del_list]
            timetable[mi] = cleaned

        
        if model_id in timetable:
            timetable[model_id].append(new)
        else:
            timetable[model_id] = [new]
        
        timetable = self.sort_and_merge(timetable)
        self.save_tb(timetable)

    def get_model_id_by_datetime(self, dt: datetime) -> str:
        """get model id given run datetime

        :param datetime dt: run datetime
        :return str: model id in place
        """
        timetable = self.load_tb()
        for mi, tis in timetable.items():
            for ti in tis:
                if dt in ti:
                    return mi
                
    def get_schedule(self) -> pd.DataFrame:
        """get runnning timetable

        :return pd.DataFrame: timetable in pandas
        """
        timetable = self.load_tb()
        
        r =[]
        for mi, tis in timetable.items():
            for ti in tis:
                t = ti.to_js()
                t['model_id'] = mi
                r.append(t)
        
        df = pd.DataFrame.from_records(r)
        if len(df) > 0:
            df['start'] = pd.to_datetime(df['start'])
            df['end'] = pd.to_datetime(df['end'])
            df.sort_values(by = 'start', inplace=True, ignore_index=True)

        return df
    
class ModelTimeTableLocalFS:
    """you can have different model (model_id) run in different time period as prod model
    support for multiple schedules
    """
    def __init__(self, tb_js_path: str) -> None:
        self.tb_js_path = tb_js_path

    def load_tb(self) -> Dict[str, List[TimeInterval]]:
        """load time table

        :return Dict[str, List[TimeInterval]]: time table loaded
        """
        tb = {}
        if os.path.exists(self.tb_js_path):
            c = loadJSON(self.tb_js_path)
            for model_id, intervals in c.items():
                tb[model_id] = [TimeInterval.from_js(interval) for interval in intervals]
        return tb
    
    def save_tb(self, tb: Dict[str, List[TimeInterval]]):
        """save timetable

        :param Dict[str, List[TimeInterval]] tb: time table
        """
        c = {}
        for model_id, intervals in tb.items():
            c[model_id] = [interval.to_js() for interval in intervals]
        toJSON(c, self.tb_js_path)
        
        
class ModelTimeTableObjStorage:
    """you can have different model (model_id) run in different time period as prod model
    support for multiple schedules
    """ 
    def __init__(self, objstore: ObjStorage, tb_js_path: str):
        """
        
        :param tb_js_path: maybe not used if not local file system
        """
        self.tb_js_path = tb_js_path
        self._objstore = objstore

    def load_tb(self) -> Dict[str, List[TimeInterval]]:
        """load time table

        :return Dict[str, List[TimeInterval]]: time table loaded
        """
        tb = {}
        c = json.loads(
            self._objstore.read_obj(self.tb_js_path), 
            object_pairs_hook = OrderedDict
        )
        for model_id, intervals in c.items():
            tb[model_id] = [TimeInterval.from_js(interval) for interval in intervals]
        return tb
    
    def save_tb(self, tb: Dict[str, List[TimeInterval]]):
        """save timetable

        :param Dict[str, List[TimeInterval]] tb: time table
        """
        c = {}
        for model_id, intervals in tb.items():
            c[model_id] = [interval.to_js() for interval in intervals]
        
        content = bytes(json.dumps(c).encode('UTF-8'))
        self._objstore.save_obj(content, self.tb_js_path)


import ibis
from ibis import _

class SimpleSamplerIbis:
    def __init__(self, shuffle_key: str, sample_frac: float = 0.25, 
                 random_seed: int = 0, verbose: bool = True):
        """down sample without any constraint

        :param str shuffle_key: id column for identitying each row, if given, will shuffle
        :param float sample_frac: down sampling size, defaults to 0.25
        :param int random_seed: control randomness
        :param bool verbose: whether to print out validation message and statistics
        """
        self.random_seed = random_seed
        self.verbose = verbose
        self.shuffle_key = shuffle_key
        if sample_frac > 1 or sample_frac <= 0:
            raise ValueError("sample frac can only between (0,1]")
        self.sample_frac = sample_frac
        
    def sample(self, dataset: ibis.expr.types.Table) -> ibis.expr.types.Table:
        """execute the sampling

        :param ibis.expr.types.Table dataset: ibis dataset
        :return ibis.expr.types.Table: sampled ibis dataset
        """
        if self.shuffle_key not in dataset.columns:
            raise KeyError(f"Shuffle key {self.shuffle_key} not found in dataset")
        if self.sample_frac == 1:
            return dataset
        
        selected = (
            dataset
            .order_by(
                (_[self.shuffle_key]
                .try_cast("string") + str(self.random_seed)) # randomness here
                .hash()
            )
            .mutate(
                (ibis.row_number().over() 
                / _.count().over() > self.sample_frac)
                .ifelse(0, 1)
                .name('IS_SELECTED_')
            )
            .filter(
                _['IS_SELECTED_'] == 1
            ).drop('IS_SELECTED_')
        )
        
        return selected


class StratifiedSamplerIbis(SimpleSamplerIbis):
    def __init__(self, shuffle_key: str, stratify_key: str, sample_frac: float = 0.25, 
                 random_seed: int = 0, verbose: bool = True):
        """down sample with roughly equal distributon of split_key before and after sampling

        :param str shuffle_key: id column for identitying each row, if given, will shuffle
        :param str stratify_key: the column name for determine stratify
        :param float sample_frac: down sampling size, defaults to 0.25
        :param int random_seed: control randomness
        :param bool verbose: whether to print out validation message and statistics
        """
        super().__init__(
            sample_frac = sample_frac,
            shuffle_key = shuffle_key,
            random_seed = random_seed,
            verbose = verbose
        )
        self.stratify_key = stratify_key

        
    def sample(self, dataset: ibis.expr.types.Table) -> ibis.expr.types.Table:
        """execute the sampling

        :param ibis.expr.types.Table dataset: ibis dataset
        :return ibis.expr.types.Table: sampled ibis dataset
        """
        if self.shuffle_key not in dataset.columns:
            raise KeyError(f"Shuffle key {self.shuffle_key} not found in dataset")
        if self.stratify_key not in dataset.columns:
            raise KeyError(f"Stratify key {self.stratify_key} not found in dataset")
        if self.sample_frac == 1:
            return dataset
        
        if self.shuffle_key:
            dataset = (
                dataset
                
            )
        
        selected = (
            dataset
            .mutate(
                (_[self.shuffle_key]
                .try_cast("string") + str(self.random_seed)) # randomness here
                .hash()
                .name('RAND_')
            )
            .mutate(
                (
                    (
                    ibis
                    .row_number()
                    .over(
                        ibis.window(
                            group_by = [self.stratify_key],
                            order_by = ['RAND_']
                        )
                    ) # partition by target column -- stratify
                    ) / (
                        _
                        .count()
                        .over(
                            ibis.window(
                                group_by = [self.stratify_key]
                            )
                        )
                    ) > self.sample_frac
                ).ifelse(0, 1)
                .name('IS_SELECTED_')
            )
        )
        return (
            selected
            .filter(selected['IS_SELECTED_'] == 1)
            .drop('IS_SELECTED_', 'RAND_')
        )


import logging
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine
from luntaiDs.CommonTools.dbapi import baseDbInf


class TeraPlaceholderEngine:
    def __init__(self, engine):
        self.engine = engine

    def connect(self):
        return self.engine

    def raw_connection(self):
        return self.engine

    def begin(self):
        return self.engine

class TeraData(baseDbInf):
    def __init__(self, driver_type:str = "sqlalchemy", driver = "teradatasql", log_mechanism: str = "LDAP"):
        """

        :param driver_type: {odbc, teradata, sqlalchemy}
        :param driver:
        :param log_mechanism: {LDAP, TD2, JWT, KRB5}

        https://pypi.org/project/teradatasqlalchemy/

        """
        super().__init__(driver)
        self.driver_type = driver_type
        self.log_mechanism = log_mechanism

    def argTempStr(self):
        return "?"

    def getConnStr(self) -> str:
        # only used for sqlalchemy driver type
        'teradatasql://{username}:{password}@{hostname}'
        return f"{self.driver}://{self.username}:{self.password}@{self.ip}"

    def launch(self):
        """Launch the databse connector, create the sqlalchemy engine and create a session

        :return:
        """
        if self.driver_type == "odbc":
            import pyodbc

            engine = pyodbc.connect(f'DSN=Teradata;'f'UID={self.username}; PWD={self.password}')
            engine.setdecoding(pyodbc.SQL_CHAR, encoding="utf-8")
            engine.setdecoding(pyodbc.SQL_WCHAR, encoding="utf-8")
            engine.setencoding(encoding="utf-8")
            self.engine = TeraPlaceholderEngine(engine)

        elif self.driver_type == "teradata":
            import teradatasql
            
            engine = teradatasql.connect(host = self.ip, user = self.username, password = self.password)
            self.engine = TeraPlaceholderEngine(engine)

        elif self.driver_type == "sqlalchemy":
            connStr = self.getConnStr()
            self.engine = create_engine(connStr)
            self.DBSession = sessionmaker(bind = self.engine)

        logging.info("Engine started, ready to go!")

    def getJdbcUrl(self) -> str:
        # for pyspark use
        return f"jdbc:teradata://{self.ip}/{self.db}"

    def getDriverClass(self) -> str:
        # for pyspark use
        return "com.teradata.jdbc.TeraDriver"


import logging
import pandas as pd
import numpy as np
import contextlib

from sqlalchemy import create_engine, and_
from sqlalchemy.orm import sessionmaker, Session, Query
from sqlalchemy.inspection import inspect
from urllib.parse import quote_plus

#######################################
###  Base Database Configuration    ###
#######################################

class baseDbInf:
    def __init__(self, driver: str):
        """the base configuration object, can be inherent by MySQL, DB2, SQLite, etc.

        :param driver: the driver to use, string value
        """
        self.driver = driver

    def bindServer(self, ip: str = None, port: int = None, db: str = None):
        """Connect to a database server

        :param ip: ip of the server
        :param port: port number of the server
        :param db: which database to connect
        :return:
        """
        self.ip = ip
        self.port = port
        self.db = db

    def login(self, username: str, password: str):
        """Login to server-client based database (MySQL, DB2, SqlServer, Hive, etc.)

        :param username:
        :param password:
        :return:
        """
        self.username = username
        self.password = password

    def argTempStr(self):
        raise NotImplementedError("Must implement this method to get different arg placeholder for different database")

    def getConnStr(self) -> str:
        """Connection string for sqlalchemy

        :return:
        """
        #engine = create_engine('mysql+mysqlconnector://USRNAME:PSWD@localhost:3306/DATABASE?charset=utf8')
        #engine = create_engine("ibm_db_sa://USRNAME:PSWD@IP:PORT/DATABASE?charset=utf8")
        #engine = create_engine('sqlite:///DB_ADDRESS')
        # https://stackoverflow.com/questions/1423804/writing-a-connection-string-when-password-contains-special-characters
        return f"{self.driver}://{self.username}:{quote_plus(self.password)}@{self.ip}:{self.port}/{self.db}?charset=utf8"

    def launch(self):
        """Launch the databse connector, create the sqlalchemy engine and create a session

        :return:
        """
        connStr = self.getConnStr()
        self.engine = create_engine(connStr)
        self.DBSession = sessionmaker(bind = self.engine)
        logging.info("Engine started, ready to go!")

    def newSession(self):
        try:
            session = self.DBSession()
        except Exception as e:
            logging.error(e)
        else:
            return session

    def getJdbcUrl(self) -> str:
        """Get JDBC connection string, for spark connection and other purpose

        :return:
        """
        raise NotImplementedError("Must implement this method to get different JdbcUrl for different database")

    def getDriverClass(self) -> str:
        # for spark connection and other purpose
        raise NotImplementedError("Must implement this method to get different DriverClass for different database")


#############################################
### Provider level database Configuration ###
#############################################

class MySQL(baseDbInf):
    def __init__(self, driver = "mysqlconnector"):
        super().__init__(f"mysql+{driver}")

    def argTempStr(self):
        return "%s"

    def getJdbcUrl(self) -> str:
        return f"jdbc:mysql://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/mysql/mysql-connector-java/8.0.27/mysql-connector-java-8.0.27.jar
        return "com.mysql.jdbc.Driver"

class DB2(baseDbInf):
    def __init__(self):
        super().__init__("ibm_db_sa")

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:db2://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/com/ibm/db2/jcc/db2jcc/db2jcc4/db2jcc-db2jcc4.jar
        return "com.ibm.db2.jcc.DB2Driver"


class SqLite(baseDbInf):
    def __init__(self, dburl):
        super().__init__("sqlite")
        self.dburl = dburl

    def getConnStr(self) -> str:
        return f"sqlite:///{self.dburl}"

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:sqlite://{self.dburl}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/org/xerial/sqlite-jdbc/3.36.0.3/sqlite-jdbc-3.36.0.3.jar
        return "org.sqlite.JDBC"


class SqlServer(baseDbInf):
    def __init__(self, instance: str):
        super().__init__("mssql+pymssql")
        self.instance = instance
        # TODO: if pymssql is missing, can consider using pyodbc way:
        #  https://datadiaries.dev/connecting-to-microsoft-sql-server-using-sqlalchemy-and-pyodbc

    def argTempStr(self):
        return "?"

    def getConnStr(self) -> str:
        # only used for sqlalchemy driver type
        # "'DRIVER={ODBC Driver 17 for SQL Server};SERVER='+server+';DATABASE='+database+';UID='+username+';PWD='+ password"
        return fr"{self.driver}://MAPLE\{self.username}:{self.password}@/{self.db}?host={self.ip}:{self.port}\{self.instance}"

    def getJdbcUrl(self) -> str:
        return fr"jdbc:sqlserver://{self.ip}\{self.instance}:{self.port}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/com/microsoft/sqlserver/mssql-jdbc/8.4.0.jre8/mssql-jdbc-8.4.0.jre8.jar
        return "com.microsoft.sqlserver.jdbc.SQLServerDriver"

class Oracle(baseDbInf):
    def __init__(self, connection_type: str = 'service_name'):
        """

        :param connection_type: {service_name, sid}
        """
        super().__init__("oracle+cx_oracle")
        self.connection_type = connection_type

    def bindServer(self, ip: str = None, port: int = None, db: str = None, sid: str = None):
        """

        :param ip:
        :param port:
        :param db: can be null
        :param sid: service name or sid
        :return:
        """
        super().bindServer(ip, port, db)
        self.sid = sid

    def argTempStr(self):
        return "?"

    def getConnStr(self) -> str:
        # only used for sqlalchemy driver type
        # "'DRIVER={ODBC Driver 17 for SQL Server};SERVER='+server+';DATABASE='+database+';UID='+username+';PWD='+ password"
        if self.connection_type == 'service_name':
            return f"{self.driver}://{self.username}:{self.password}@{self.ip}:{self.port}/?service_name={self.sid}"
        else:
            return f"{self.driver}://{self.username}:{self.password}@{self.ip}:{self.port}/{self.sid}"

    def getJdbcUrl(self) -> str:
        if self.connection_type == 'service_name':
            return f"jdbc:oracle:thin:@//{self.ip}:{self.port}/{self.sid}"
        else:
            return f"jdbc:oracle:thin:@{self.ip}:{self.port}:{self.sid}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/com/oracle/database/jdbc/ojdbc8/19.3.0.0/ojdbc8-19.3.0.0.jar
        return "oracle.jdbc.driver.OracleDriver"

class PostgresSql(baseDbInf):
    def __init__(self):
        super().__init__("postgresql+psycopg2")

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:postgresql://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/org/postgresql/postgresql/42.2.18/postgresql-42.2.18.jar
        return "org.postgresql.Driver"


"""-------------------------------------------------------------------------------------------------"""

#######################################
### Database accesor Object -- DAO  ###
#######################################

class dbIO:

    def __init__(self, dbinf: baseDbInf):
        """database accessor designed for pandas and sqlalchemy

        :param dbinf: database configuration object (DB2, MySQL, etc.)
        """
        dbinf.launch()
        self.__dbinf = dbinf
        self.placeholder = dbinf.argTempStr()

    def getDbInf(self):
        return self.__dbinf

    def getEngine(self):
        return self.getDbInf().engine

    @ contextlib.contextmanager
    def get_session(self, errormsg : str = "reason Unknown") -> Session:
        """return the session object to operate

        :param errormsg: the error message to be displayed if error occured
        :return: Session object

        example:
        >>> with self.get_session() as s:
        >>>     #do something here
        """
        session = self.getDbInf().newSession()
        try:
            yield session
            session.commit()
        except Exception as e:
            session.rollback()
            logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
        finally:
            session.close()

    def insert(self, table, record : dict) -> None:
        """insert a record into db

        :param table: table class
        :param record: dict of record to insert
        :return: None

        example:
        >>> x = {"fund_id" : "160023", "date" : "2005-09-02", "net_value" : 1.0102, "full_value" : 3.0102, "div" : 0.234, "pnl" : 0.2341}
        >>> dbIO.insert(MutualFundHist, x)
        """
        with self.get_session() as s:
            s.add(table(**record))

    def update(self, table, primary_kvs : dict, record : dict) -> None:
        """update record by looking at primary key

        :param table: table class
        :param primary_kvs: dict of primary key-value pairs, use to find which record(s) to update
        :param record: dict of new record
        :return: None

        example:
        >>> p = {"fund_id" : "150234", "date" : "2005-09-09"}  # use to find which records to update_all
        >>> r = {"net_value" : 0.4456}  # the new record to save to db
        >>> dbIO.update(MutualFundHist, p, r)
        is equivalent to:
        UPDATE MutualFundHist
        SET net_value = 0.4456
        WHERE fund_id = '150234' AND date = '2005-09-09';
        """
        with self.get_session() as s:
            #s.query_extract(table).filter(table.fund_id == "150234").update_all(r)
            conditions = [getattr(table, k) == v for k, v in primary_kvs.items()]
            s.query(table).filter(and_(*conditions)).update_all(record)

    def delete(self, table, primary_kvs : dict) -> None:
        """delete record by looking at primary key

        :param table: table class
        :param primary_kvs: dict of primary key-value pairs, use to find which record(s) to update_all
        :return: None

        example:
        >>> p = {"fund_id" : "150234", "date" : "2005-09-09"}  # use to find which records to update_all
        >>> self.delete(MutualFundHist, p)
        is equivalent to:
        DELETE FROM MutualFundHist
        WHERE fund_id = '150234' AND date = '2005-09-09';
        """
        with self.get_session() as s:
            conditions = [getattr(table, k) == v for k, v in primary_kvs.items()]
            s.query(table).filter(and_(*conditions)).delete()


    def modify_sql(self, sql : str, *args, errormsg : str = "reason Unknown", **kws) -> int:
        """execute original sql (not query_extract)

        :param sql: sql string
        :param args: positional arguments to fill in sql template
        :param kws: kw arguments to fill in sql template
        :param errormsg: error message to display when error occurred
        :return: 1: success, 0: failed

        example:
        >>> sql = 'UPDATE MutualFundHist SET net_value = %s WHERE fund_id = %s AND date = %s;'
        >>> self.modify_sql(sql, 0.7456, '150234', '2005-09-09')
        """
        with self.getEngine().connect() as conn:
            try:
                sql = sql.replace("?", self.placeholder)
                conn.execute(sql, *args, **kws)
            except Exception as e:
                logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
                return 0
            else:
                return 1

    def query_df(self, query : Query) -> pd.DataFrame:
        """making query_extract directly

        :param query: Query statement
        :return: resulting dataframe

        example:
        >>> with self.get_session() as s:
        >>>     query_extract = s.query_extract(MutualFundHist).filter(MutualFundHist.fund_id == "160023")
        >>>
        >>> df = self.query_df(query_extract)
        """
        return pd.read_sql(query.statement, query.session.bind).replace({None : np.nan})

    def query_sql_df(self, sql : str, *args, errormsg : str = "reason Unknown", **kws) -> pd.DataFrame:
        """using original sql to make queries

        :param sql: the sql template
        :param args: the argument list to fill in the sql template
        :param errormsg: message to display when encoutering errors
        :param kws: the argument kws to fill in the sql template
        :return: the result dataframe

        example:
        >>> sql = "select * from mutualfundhist where fund_id = ?"
        >>> r = self.query_sql_df(sql, '160023')
        """
        with self.getEngine().connect() as conn:
            try:
                sql = sql.replace("?", self.placeholder)
                r = conn.execute(sql, *args, **kws)
                #df = pd.read_sql(sql, conn, params = args)
            except Exception as e:
                logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
                df = pd.DataFrame({})
            else:
                if hasattr(r, "cursor"):
                    description = r.cursor.description
                else:
                    description = r.description

                headers = [ i[0] for i in description ]
                # df = pd.DataFrame.from_records(r, columns = headers)
                df = pd.DataFrame.from_records(r, columns = headers, coerce_float = True)
            finally:
                return df

    def insert_df(self, table, df : pd.DataFrame) -> None:
        """save a pandas table into database

        :param table: table class
        :param df: the pandas Dataframe to save into db
        :return:

        example:
        >>> df = pd.DataFrame({
        >>>     "fund_id" : ["160023", "160023", "160023", "150234"],
        >>>     "date" : ["2005-09-09", "2005-09-02", "2005-08-31", "2005-09-09"],
        >>>     "net_value" : [1.0234, 1.0102, 1.0456, 0.9876],
        >>>     "full_value" : [3.0234, 3.0102, 3.0456, 2.9876],
        >>>     "div" : [None, 0.234, None, None],
        >>>     "split_ratio" : [1.0034, None, None, 1.2232],
        >>>     "pnl" : [0.2334, 0.2341, -0.1442, -0.0032],
        >>> })
        >>> self.insert_df(MutualFundHist, df)
        """
        records = df2Tables(df, table)

        with self.get_session() as s:
            s.add_all(records)

    def insert_pd_df(self, tablename: str, df : pd.DataFrame, schema: str = None, mode: str = "append", index: bool = False,
                     chunksize: int = 1000, method: str = 'multi', errormsg : str = "reason Unknown") -> int:
        """save a pandas dataframe to database using pandas style

        :param tablename: the name str of the table
        :param df: the pandas dataframe to insert
        :param schema: the schema you are going to drop table into
        :param mode: {‘fail’, ‘replace’, ‘append’}, default ‘append’
        :param index: whether to include the index
        :param chunksize: Specify the number of rows in each batch to be written at a time. recommend 1000 or None
        :param method: {None, ‘multi’, callable}
        :return: 1: success, 0: failed

        check here: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_sql.html
        speed up insert:
        use method = multi will significantly speed up as it will insert in a batch
        limit chunk size in accordance with the limit of your database
        """
        try:
            with self.getEngine().begin() as conn:  # .begin() will automatically tackle with rollback() and commit()
                df.to_sql(tablename, conn, schema = schema, if_exists = mode, index = index, chunksize = chunksize, method = method)
        except Exception as e:
            logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
            return 0
        else:
            return 1


    def insert_sql_df(self, tablename: str, df : pd.DataFrame, errormsg : str = "reason Unknown") -> int:
        """save a pandas dataframe to database using pythonic style

        :param tablename: the name str of the table
        :param df: the pandas dataframe to insert
        :return: 1: success, 0: failed
        """
        sql = f"""
        insert into {tablename} ({",".join(df.columns)})
        values ({",".join("?" for col in df.columns)})
        """
        sql = sql.replace("?", self.placeholder)

        conn = self.getEngine().raw_connection()
        cursor = conn.cursor()
        try:
            cursor.executemany(sql, df.values)
        except Exception as e:
            conn.rollback()
            logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
            return 0
        else:
            conn.commit()
            return 1
        finally:
            cursor.close()
            conn.close()



def getPrimaryKeys(table) -> list:
    return [key.name for key in inspect(table).primary_key]

def df2Tables(df : pd.DataFrame, table) -> list:
    records = []
    for row in df.replace({np.nan: None}).to_dict(orient = "records"):
        records.append(table(**row))

    return records

#######################################
### JDBC drivers download and setup ###
#######################################


if __name__ == '__main__':


    pass


from typing import List, Dict
from datetime import datetime
import pandas as pd
import pymongo
from pymongo.mongo_client import MongoClient
from luntaiDs.ModelingTools.Serving.model_registry import _BaseModelRegistry
from luntaiDs.ModelingTools.Serving.timetable import _BaseModelTimeTable, TimeInterval
from luntaiDs.ProviderTools.mongo.serving import MongoClient

class _BaseModelRegistryMongo(_BaseModelRegistry):
    """Base class for Simple Model Registry on MongoDB
    Note Mongo only means the configuration will be saved in MongoDB, not model data
    model versioning is achieved by tracking model_id and its configuration

    Different from its super class, each model config will be a seperate entry in one collection (table)
    and prod key is added to each entry
    each model configuration format:
    {
        "model_id" : "MODEL_ID_A",
        "is_prod" : true,
        "config" : {CONFIG_A}
    }
    {
        "model_id" : "MODEL_ID_B",
        "is_prod" : false,
        "config" : {CONFIG_B}
    }
    """
    def __init__(self, mongo_client: MongoClient, db: str, collection: str):
        """
        
        :param mongo_client: the mongo client to connect with
        :param db: the database name
        :param collection: the collection (table) name for record
        """
        self._mongo = mongo_client
        self._collection = self._mongo[db][collection]
        self.db = db
        self.collection = collection
        
    def load_config(self) -> dict:
        """load configuration
        
        note that it will combine all config into 1 dictionary and convert to base format
        configuration format:
        {
            "prod" : "MODEL_ID",
            "archive" : {
                "MODEL_ID_A" : {CONFIG_A},
                "MODEL_ID_B" : {CONFIG_B},
            }
        }
        
        :return dict: configuration in dictionary format
        """
        prod_model_id = None
        archives =  {}
        for r in self._collection.find(
            {}, # find all
            {'_id': 0}, # drop id column
            sort = [( '_id', pymongo.DESCENDING )] # sort by latest record
            
        ):
            model_id = r['model_id']
            if r['is_prod']:
                prod_model_id = model_id
            archives[model_id]= r['config']

        return {
            "prod" : prod_model_id,
            "archive" : archives
        }
        
    def save_config(self, config:dict):
        """save configuration dictionary, overwrite mode
        
        note that the config here is the base format
        configuration format:
        {
            "prod" : "MODEL_ID",
            "archive" : {
                "MODEL_ID_A" : {CONFIG_A},
                "MODEL_ID_B" : {CONFIG_B},
            }
        }

        :param dict config: configuration in base dictionary format
        """
        prod_model_id = config.get('prod')
        model_ids = []
        for model_id, model_config in config.get('archive', {}).items():
            model_ids.append(model_id)
            is_prod = False
            if model_id == prod_model_id:
                is_prod = True
                
            # construct content
            content = {
                "model_id" : model_id,
                "is_prod" : is_prod,
                "config" : model_config
            }
            self._collection.replace_one(
                filter = {
                    'model_id': model_id
                }, # find matching record, if any
                replacement = content,
                upsert = True # update if found, insert if not found
            )
        
        # delete model_id that are not in config
        self._collection.delete_many(
            { 'model_id' :  { '$nin': model_ids }}
        )
        

    def get_prod_model_id(self) -> str:
        """get deployed (prod) model id

        :return str: model id that is deployed
        """
        record = (
            self._collection
            .find_one(
                {'is_prod': True}, # matching condition
                {'_id': 0, 'model_id' : 1}, # drop id column
                sort = [( '_id', pymongo.DESCENDING )] # in case multiple, find the latest record
            )
        )
        if record is None:
            return None
        return record.get('model_id')
    
    def get_model_list(self) -> List[str]:
        """get a list of model ids

        :return List[str]: list of model ids registered
        """
        return self._collection.distinct('model_id')
    
    def get_model_config(self, model_id: str) -> dict:
        """get specific configuration part for a given model id

        :param str model_id: model id of the model
        :return dict: the specific configuration part
        """
        assert model_id in self.get_model_list(), "Model Id not found in registry, please register one first"
        record = (
            self._collection
            .find_one(
                {'model_id': model_id}, # matching condition
                {'_id': 0, 'config' : 1}, # drop id column
                sort = [( '_id', pymongo.DESCENDING )] # in case multiple, find the latest record
            )
        )
        return record.get('config')
        
    def upsert_model_config(self, model_id: str, model_config: dict):
        """update or insert (if not exist) model configuration

        :param str model_id: model id of the model
        :param dict model_config: the specific configuration part
        """
        is_prod_field = (
            self._collection
            .find_one(
                {'model_id': model_id}, # matching condition
                {'_id': 0, 'is_prod' : 1}, # drop id column
                sort = [( '_id', pymongo.DESCENDING )] # in case multiple, find the latest record
            )
        )
        if is_prod_field is None:
            is_prod = False
        else:
            is_prod = is_prod_field.get('is_prod', False)
            
        # construct content
        content = {
            "model_id" : model_id,
            "is_prod" : is_prod,
            "config" : model_config
        }
        self._collection.replace_one(
            filter = {
                'model_id': model_id
            }, # find matching record, if any
            replacement = content,
            upsert = True # update if found, insert if not found
        )
        
    def delete_model_config(self, model_id: str):
        """delete model configuration

        :param str model_id: model id of the model
        """
        self._collection.delete_one({
            'model_id' : model_id
        })

    def deploy(self, model_id: str):
        """deploy selected model to prod, simply achieved by setting it to prod field

        :param str model_id: model id to be deployed
        """
        # check if model id exists
        assert model_id in self.get_model_list(), "Model Id not found in registry, please register one first"
        # set all existing prod to False
        self._collection.update_many(
            {
                'is_prod': True
            }, # filter: find matching record, if any
            { 
                "$set": { 'is_prod': False } 
            } # update content 
        )
        # set selected model id to have True prod state
        self._collection.update_one(
            {
                'model_id': model_id
            }, # filter: find matching record, if any
            { 
                "$set": { 'is_prod': True } 
            } # update content 
        )
        
        

class ModelTimeTableMongo(_BaseModelTimeTable):
    def __init__(self, mongo_client: MongoClient, db: str, collection: str):
        self._mongo = mongo_client
        self._collection = self._mongo[db][collection]
        
    def load_tb(self) -> Dict[str, List[TimeInterval]]:
        """load time table

        :return Dict[str, List[TimeInterval]]: time table loaded
        """
        result = {}
        for record in self._collection.find({}):
            model_id = record.get('model_id')
            ti = TimeInterval(
                start = record.get('start'), 
                end = record.get('end')
            )
            if model_id in result:
                result[model_id].append(ti)
            else:
                result[model_id] = [ti]
        return result
                
    
    def save_tb(self, tb: Dict[str, List[TimeInterval]]):
        """save timetable

        :param Dict[str, List[TimeInterval]] tb: time table
        """
        # need to overwrite everything, so need to delete all first
        self._collection.delete_many({})
        
        for model_id, intervals in tb.items():
            for interval in intervals:
                self._collection.insert_one(
                    {
                        'model_id' : model_id,
                        'start' : interval.start,
                        'end' : interval.end
                    }
                )
                
    def get_model_id_by_datetime(self, dt: datetime) -> str:
        """get model id given run datetime

        :param datetime dt: run datetime
        :return str: model id in place
        """
        record = (
            self._collection
            .find_one(
                {'start': {'$lte' : dt}, 'end' : {'$gte' : dt}}, # matching condition
                {'_id': 0}, # drop id column
                sort = [( '_id', pymongo.DESCENDING )] # in case multiple, find the latest record
            )
        )
        if record is None:
            return None
        return record.get('model_id')
    
    def get_schedule(self) -> pd.DataFrame:
        """get runnning timetable

        :return pd.DataFrame: timetable in pandas
        """
        return pd.DataFrame(
            self._collection
            .find({}, { "_id": 0})
            .sort('start')
        )


import logging
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine
import ibis
import pandas as pd
from google.oauth2 import service_account
from luntaiDs.CommonTools.dbapi import baseDbInf
from luntaiDs.CommonTools.dtyper import DSchema
from luntaiDs.CommonTools.warehouse import BaseWarehouseHandler

class BigQuery(baseDbInf):
    def __init__(self, project_id: str, driver="bigquery", credentials_path: str = None):
        """
        bq = BigQuery(project_id='your-project-id')
        bq.bindServer(db = 'your-dataset-id')
        bq.launch()
        """
        super().__init__(driver)
        self.project_id = project_id
        self.credentials_path = credentials_path # json file
        self.bindServer(db = None)

    def argTempStr(self):
        return "%s"

    def getJdbcUrl(self) -> str:
        return f"jdbc:bigquery://{self.ip}:{self.port};ProjectId={self.project_id}"

    def getDriverClass(self) -> str:
        # https://mvnrepository.com/artifact/com.google.cloud/google-cloud-bigquery/2.37.1
        return "com.simba.googlebigquery.jdbc42.Driver"

    def getConnStr(self) -> str:
        # only used for sqlalchemy driver type
        return f"{self.driver}://{self.project_id}/{self.db}"
    
    def bindServer(self, ip: str = 'https://www.googleapis.com/bigquery/v2', port: int = 443, db: str = None):
        """Connect to a database server

        :param ip: ip of the server
        :param port: port number of the server
        :param db: which database to connect
        :return:
        """
        self.ip = ip
        self.port = port
        self.db = db

    def launch(self):
        """Launch the databse connector, create the sqlalchemy engine and create a session

        :return:
        """
        connStr = self.getConnStr()
        self.engine = create_engine(connStr, credentials_path=self.credentials_path)
        self.DBSession = sessionmaker(bind = self.engine)
        logging.info("Engine started, ready to go!")
        

class WarehouseHandlerBQSQL(BaseWarehouseHandler):
    @classmethod
    def connect(cls, db_conf: BigQuery, **settings):
        cls._db_conf = db_conf
        if db_conf.credentials_path:
            creds = service_account.Credentials.from_service_account_file(db_conf.credentials_path)
        else:
            creds = None
        cls._ops = ibis.bigquery.connect(
            project_id=db_conf.project_id,
            dataset_id=db_conf.db,
            credentials = creds if creds else None,
            **settings
        )
        
    def is_exist(self, schema: str, table: str) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created
        
        :param str schema: schema/database
        :param str table: table name
        :return bool: whether the table and schema exists and ready
        """
        if schema not in self._ops.list_schemas(like = schema):
            return False
        return table in self._ops.list_tables(like = table, schema = schema)
    
    def get_table(self, schema: str, table: str) -> ibis.expr.types.Table:
        """get the ibis table
        
        :param str schema: schema/database
        :param str table: table name
        """
        return self._ops.table(name = table, schema = schema)
    
    def create_schema(self, schema: str):
        """create database/schema
        
        :param str schema: schema/database
        """
        self._ops.create_schema(
            name = schema,
            force = True
        )
        
    def create_table(self, schema: str, table: str, col_schemas: DSchema, **settings):
        """initialize/create table in the underlying data warehouse system

        :param str schema: schema/database
        :param str table: table name
        :param DSchema col_schemas: data column schema
        :param List[str] primary_keys: primary keys, defaults to None
        """
        # create table
        primary_keys = col_schemas.primary_keys
        partition_keys = col_schemas.partition_keys
        cluster_keys = col_schemas.cluster_keys
        logging.info(f"Creating table {schema}.{table} using schema:\n{col_schemas.ibis_schema}")
        logging.info(f"primary keys = {primary_keys}, partition_keys = {partition_keys}")
        # create table
        self._ops.create_table(
            name = table,
            schema = col_schemas.ibis_schema,
            database = schema,
            overwrite = False,
            partition_by = partition_keys,
            cluster_by = cluster_keys,
            options = settings
        )
        # add column descriptions
        descrs = col_schemas.descrs
        for col, descr in descrs.items():
            try:
                sql = f"""
                ALTER TABLE {schema}.{table} 
                ALTER COLUMN {col} 
                SET OPTIONS (
                    description="{descr}"
                )"""
                self._ops.con.command(sql)
            except:
                pass
            
    def get_dtypes(self, schema: str, table: str) -> ibis.Schema:
        """Check dtypes of the given schema/dataset

        :return:
        """
        return self._ops.get_schema(
            name = table,
            schema = schema
        )
        
    def query(self, sql: str, schema: str = None) -> ibis.expr.types.Table:
        """query using SQL

        :return: ibis dataframe
        """
        return self._ops.sql(sql)
    
    def execute(self, sql: str, **kws):
        """execute SQL for operations like insert/update/delete

        :return: ibis dataframe
        """
        self._ops.raw_sql(sql, **kws)
        
    def save_pandas(self, df: pd.DataFrame, schema: str, table: str, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param str schema: schema/database
        :param str table: table name
        """
        self._ops.client.load_table_from_dataframe(
            dataframe = df,
            destination = f"{schema}.{table}",
            **kws
        )
        
    def delete_table(self, schema: str, table: str):
        """drop whole table
        
        :param str schema: schema/database
        :param str table: table name
        """
        self._ops.drop_table(
            name = table,
            schema = schema,
            force = True
        )
    
    def truncate_table(self, schema: str, table: str):
        """Truncate (remove) data from table
        
        :param str schema: schema/database
        :param str table: table name
        """
        self._ops.truncate_table(
            name = table,
            schema = schema
        )


import os
import shutil
from typing import List, Dict, Union
import ibis
import pandas as pd

class _BaseModelDataRegistry:

    def get_existing_ids(self) -> List[str]:
        """get list of registered data ids

        :return List[str]: data ids
        """
        raise NotImplementedError("")

    def register(self, data_id: str, train_ds: Union[ibis.expr.types.Table | pd.DataFrame], 
                 test_ds: Union[ibis.expr.types.Table | pd.DataFrame], replace: bool = False):
        """register table in ibis format

        :param str data_id: data id
        :param Union[ibis.expr.types.Table | pd.DataFrame] train_ds: training dataset
        :param Union[ibis.expr.types.Table | pd.DataFrame] test_ds: testing dataset
        :param bool replace: whether to replace existing dataset, defaults to False
        """
        raise NotImplementedError("")

    def fetch(self, data_id: str, target_col: str = None):
        """fetch training/testing dataset

        :param str data_id: data id to be fetched
        :param str target_col: the target column, defaults to None.
        :return:
            - if target_col given, will split to [X_train, y_train, X_test, y_test]
            - if target_col not given, will just split to [train_ds, test_ds]
        """
        raise NotImplementedError("")

    def remove(self, data_id: str):
        """remove dataset from registry

        :param str data_id: data id to be removed
        """
        raise NotImplementedError("")


class ModelDataRegistryLocalFS(_BaseModelDataRegistry):
    def __init__(self, data_root: str):
        """
        data_root
        - data_id
            - train.parquet
            - test.parquet
        """
        self.data_root = data_root
        os.makedirs(data_root, exist_ok=True)

    def get_existing_ids(self) -> List[str]:
        """get list of registered data ids

        :return List[str]: data ids
        """
        return os.listdir(self.data_root)

    def get_train_path(self, data_id: str) -> str:
        return os.path.join(self.data_root, data_id, f"train_{data_id}.parquet")

    def get_test_path(self, data_id: str) -> str:
        return os.path.join(self.data_root, data_id, f"test_{data_id}.parquet")

    def register(self, data_id: str, train_ds: ibis.expr.types.Table, test_ds: ibis.expr.types.Table, replace: bool = False):
        """register table in ibis format

        :param str data_id: data id
        :param ibis.expr.types.Table train_ds: training dataset
        :param ibis.expr.types.Table test_ds: testing dataset
        :param bool replace: whether to replace existing dataset, defaults to False
        """
        existing_ids = self.get_existing_ids()
        if not replace and data_id in existing_ids:
            raise ValueError(f"data id {data_id} already exist, pls use another id")
        os.makedirs(os.path.join(self.data_root, data_id), exist_ok=False)
        train_ds.to_parquet(self.get_train_path(data_id))
        test_ds.to_parquet(self.get_test_path(data_id))

    def fetch(self, data_id: str, target_col: str = None):
        """fetch training/testing dataset

        :param str data_id: data id to be fetched
        :param str target_col: the target column, defaults to None.
        :return:
            - if target_col given, will split to [X_train, y_train, X_test, y_test]
            - if target_col not given, will just split to [train_ds, test_ds]
        """
        assert data_id in self.get_existing_ids(), f"data id {data_id} does not exist"
        train_ds = ibis.read_parquet(self.get_train_path(data_id))
        test_ds = ibis.read_parquet(self.get_test_path(data_id))
        if target_col:
            X_train = train_ds.drop(columns=[target_col])
            X_test = test_ds.drop(columns=[target_col])
            y_train = train_ds[target_col]
            y_test = test_ds[target_col]
            return X_train, y_train, X_test, y_test
        else:
            return train_ds, test_ds

    def remove(self, data_id: str):
        """remove dataset from registry

        :param str data_id: data id to be removed
        """
        dpath = os.path.join(self.data_root, data_id)
        shutil.rmtree(dpath)



from typing import List, Dict, Union
import ibis
import pandas as pd
import logging
from luntaiDs.ModelingTools.Serving.data_registry import _BaseModelDataRegistry
from luntaiDs.ProviderTools.gcp.dbapi import WarehouseHandlerBQSQL


class _BaseModelDataRegistryBQ(_BaseModelDataRegistry):
    DATA_ID_COL = "DATA_ID_"
    TRAIN_TEST_IND_COL = "IS_TRAIN_"
    
    def __init__(self, handler: WarehouseHandlerBQSQL, schema: str, table: str):
        """initialize

        :param WarehouseHandlerBQSQL handler: bigquery handler
        :param str schema: schema for the table storing modeling data
        :param str table: table storing modeling data
        """
        self.handler = handler
        self.schema = schema
        self.table = table
        
    def init_table(self):
        raise NotImplementedError("")
    
    def get_table(self) -> ibis.expr.types.Table:
        """get underlying training table using ibis

        :return ibis.expr.types.Table: ibis table
        """
        return self.handler.get_table(schema = self.schema, table = self.table)

    def get_existing_ids(self) -> List[str]:
        """get list of registered data ids

        :return List[str]: data ids
        """
        return (
            self.get_table()
            .select(self.DATA_ID_COL)
            .distinct()
            .to_pandas()
            [self.DATA_ID_COL]
            .tolist()
        )

    def register(self, data_id: str, train_ds: Union[ibis.expr.types.Table | pd.DataFrame], 
                 test_ds: Union[ibis.expr.types.Table | pd.DataFrame], replace: bool = False):
        """register table in ibis format

        :param str data_id: data id
        :param Union[ibis.expr.types.Table | pd.DataFrame] train_ds: training dataset
        :param Union[ibis.expr.types.Table | pd.DataFrame] test_ds: testing dataset
        :param bool replace: whether to replace existing dataset, defaults to False
        """
        if replace:
            self.remove(data_id = data_id)
            
        if isinstance(train_ds, pd.DataFrame) and isinstance(test_ds, pd.DataFrame):
            # add data_id column
            train_ds.loc[:, self.DATA_ID_COL] = data_id
            test_ds.loc[:, self.DATA_ID_COL] = data_id
            # add train test indicator
            train_ds.loc[:, self.TRAIN_TEST_IND_COL] = True
            test_ds.loc[:, self.TRAIN_TEST_IND_COL] = False
            # save to database
            self.handler.save_pandas(
                df = train_ds,
                schema = self.schema,
                table = self.table
            )
            self.handler.save_pandas(
                df = test_ds,
                schema = self.schema,
                table = self.table
            )
        elif isinstance(train_ds, ibis.expr.types.Table) and isinstance(test_ds, ibis.expr.types.Table):
            # add data_id column
            train_ds = (
                train_ds
                .mutate(ibis.literal(data_id, 'String').name(self.DATA_ID_COL))
                .mutate(ibis.literal(True, 'Boolean').name(self.TRAIN_TEST_IND_COL))
            )
            test_ds = (
                train_ds
                .mutate(ibis.literal(data_id, 'String').name(self.DATA_ID_COL))
                .mutate(ibis.literal(False, 'Boolean').name(self.TRAIN_TEST_IND_COL))
            )
            # save training set to database
            query_train = ibis.to_sql(train_ds)
            qry_cols_train = self.handler.query(query_train).columns
            sql_train = f"""
            INSERT {self.schema}.{self.table} ({','.join(qry_cols_train)})
            {query_train}
            """
            logging.info(f"Inserting into {self.schema}.{self.table} using query:\n{sql_train}")
            self.handler.execute(sql_train)
            # save testing set to database
            query_test = ibis.to_sql(test_ds)
            qry_cols_test = self.handler.query(query_test).columns
            sql_test = f"""
            INSERT {self.schema}.{self.table} ({','.join(qry_cols_test)})
            {query_test}
            """
            logging.info(f"Inserting into {self.schema}.{self.table} using query:\n{sql_test}")
            self.handler.execute(query_test)    
            
        else:
            raise ValueError("train_ds and test_ds must be of type either pandas or ibis dataframe")
            

    def fetch(self, data_id: str, target_col: str = None):
        """fetch training/testing dataset

        :param str data_id: data id to be fetched
        :param str target_col: the target column, defaults to None.
        :return:
            - if target_col given, will split to [X_train, y_train, X_test, y_test]
            - if target_col not given, will just split to [train_ds, test_ds]
        """
        table = self.get_table()
        train_ds = (
            table
            .filter((table[self.DATA_ID_COL] == data_id) & (table[self.TRAIN_TEST_IND_COL] == True))
            .drop(self.DATA_ID_COL, self.TRAIN_TEST_IND_COL)
        )
        test_ds = (
            table
            .filter((table[self.DATA_ID_COL] == data_id) & (table[self.TRAIN_TEST_IND_COL] == False))
            .drop(self.DATA_ID_COL, self.TRAIN_TEST_IND_COL)
        )
        if target_col:
            X_train = train_ds.drop(target_col)
            X_test = test_ds.drop(target_col)
            y_train = train_ds[target_col]
            y_test = test_ds[target_col]
            return X_train, y_train, X_test, y_test
        else:
            return train_ds, test_ds

    def remove(self, data_id: str):
        """remove dataset from registry

        :param str data_id: data id to be removed
        """
        sql = f"""
        DELETE {self.schema}.{self.table}
        WHERE {self.DATA_ID_COL} = '{data_id}'
        """
        logging.info(f"Deleting table {self.schema}.{self.table} using query:\n{sql}")
        self.handler.execute(sql)


import os
import json
from typing import List, Dict
from collections import OrderedDict
from luntaiDs.ModelingTools.Serving.model_registry import _BaseModelRegistry
from luntaiDs.ModelingTools.Serving.timetable import TimeInterval
from luntaiDs.ProviderTools.ssh.sftp import SFTP

class _BaseModelRegistrySFTP(_BaseModelRegistry):
    """Base class for Simple Model Registry on SFTP Filesystem
    Note SFTP only means the configuration will be saved in SFTP system, not model data
    model versioning is achieved by tracking model_id and its configuration

    configuration format:
    {
        "prod" : "MODEL_ID",
        "archive" : {
            "MODEL_ID_A" : {CONFIG_A},
            "MODEL_ID_B" : {CONFIG_B},
        }
    }
    """
    def __init__(self, sftp: SFTP, config_js_path: str):
        """
        
        :param config_js_path: maybe not used if not local file system
        """
        self.config_js_path = config_js_path
        self._sftp = sftp
        self._sftp.mkdir(os.path.dirname(config_js_path), ignore_existing = True)

    def load_config(self) -> dict:
        """load configuration
        
        :return dict: configuration in dictionary format
        """
        if self._sftp.exist(self.config_js_path):
            with self._sftp.getFileHandler(self.config_js_path) as obj:
                config = json.load(obj, object_pairs_hook = OrderedDict)
        else:
            config = {
                "prod" : None,
                "archive": {}
            }
        return config

    def save_config(self, config:dict):
        """save configuration dictionary

        :param dict config: configuration in dictionary format
        """
        with self._sftp.getFileHandler(self.config_js_path, 'w') as obj:
            json.dump(config, obj, indent = 4)
            
            
class ModelTimeTableSFTP:
    """you can have different model (model_id) run in different time period as prod model
    support for multiple schedules
    """
    def __init__(self, sftp: SFTP, tb_js_path: str):
        """
        
        :param tb_js_path: maybe not used if not local file system
        """
        self.tb_js_path = tb_js_path
        self._sftp = sftp
        self._sftp.mkdir(os.path.dirname(tb_js_path), ignore_existing = True)

    def load_tb(self) -> Dict[str, List[TimeInterval]]:
        """load time table

        :return Dict[str, List[TimeInterval]]: time table loaded
        """
        tb = {}
        if self._sftp.exist(self.tb_js_path):
            with self._sftp.getFileHandler(self.tb_js_path) as obj:
                c = json.load(obj, object_pairs_hook = OrderedDict)
            
            for model_id, intervals in c.items():
                tb[model_id] = [TimeInterval.from_js(interval) for interval in intervals]
        return tb
    
    def save_tb(self, tb: Dict[str, List[TimeInterval]]):
        """save timetable

        :param Dict[str, List[TimeInterval]] tb: time table
        """
        c = {}
        for model_id, intervals in tb.items():
            c[model_id] = [interval.to_js() for interval in intervals]

        with self._sftp.getFileHandler(self.tb_js_path, 'w') as obj:
            json.dump(c, obj, indent = 4)


from __future__ import annotations
import os
from datetime import date
from typing import List
import pandas as pd
from tqdm import tqdm
from luntaiDs.CommonTools.utils import str2dt
from luntaiDs.CommonTools.SnapStructure.tools import match_date_from_str
from luntaiDs.CommonTools.SnapStructure.structure import SnapshotDataManagerBase
from luntaiDs.CommonTools.sparker import SparkConnector
from luntaiDs.ProviderTools.ssh.sftp import SFTP



class SnapshotDataManagerSFTP(SnapshotDataManagerBase):
    """files are saved as parquet snapshots under each schema.table
        file naming convention: dir/tablename_YYYY-MM-DD.parquet
    """
    
    @classmethod
    def setup(cls, root_dir: str, sftp: SFTP, spark_connector: SparkConnector = None, default_engine:str = 'pandas'):
        super(SnapshotDataManagerSFTP, cls).setup(
            spark_connector = spark_connector,
            default_engine = default_engine
        )
        cls.ROOT_DIR = root_dir
        cls.default_engine = default_engine
        cls.sftp = sftp
        
    def __init__(self, schema:str, table:str):
        """virtual database management interface

        :param schema: the virtual schema , e.g., RAW, PROCESSED
        :param table:  the virtual table under each schema
        """
        super().__init__(schema = schema, table = table)
        dir = os.path.join(self.ROOT_DIR, schema, table)
        self.dir = dir  # the root path of the table under each schema
        self.init_table()
        
    def exist(self) -> bool:
        """whether the schema and table exist, or ready to do operations
        for DB based, usually it detects whether the table shema structure is created

        :return bool: whether the table and schema exists and ready
        """
        return self.sftp.exist(self.dir)

    def init_table(self):
        self.sftp.mkdir(path = os.path.join(self.ROOT_DIR, self.schema), mode = 511, ignore_existing = True)
        self.sftp.mkdir(path = self.dir, mode = 511, ignore_existing = True)
        
    def get_filename(self, snap_dt: date) -> str:
        return f"{self.table}_{snap_dt}.parquet"

    def get_default_file_path(self, snap_dt: date) -> str:
        return os.path.join(self.dir, self.get_filename(snap_dt))
    
    def _save(self, df: pd.DataFrame, snap_dt: date, **kws):
        """The pure logic to save pandas dataframe to the system, without handling existing record problem

        :param pd.DataFrame df: _description_
        :param date snap_dt: _description_
        :raises NotImplementedError: _description_
        """
        filepath = self.get_default_file_path(snap_dt)
        with self.sftp.getFileHandler(filepath, mode = 'w') as fp:
            df.to_parquet(fp, **kws)
         
    
    def get_existing_snap_dts(self) -> List[date]:
        file_list = self.sftp.ls(
            remoteFolderPath = self.dir,
            pattern = "*.parquet",
            verbose = False
        )
        existing_snaps = [
            str2dt(match_date_from_str(f))
            for f in file_list
        ]
        existing_snaps.sort()
        return existing_snaps
    
    def read(self, snap_dt: date, **kws) -> pd.DataFrame:
        """Read as pandas dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_default_file_path(snap_dt=snap_dt)
        with self.sftp.getFileHandler(filepath) as fp:
            df = pd.read_parquet(fp, **kws)
        return df
    
    def reads(self, snap_dts: List[date], partition_key: str = None, **kws) -> pd.DataFrame:
        """reads as pandas dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :param partition_key: when vertically combining, need partition key to tell apart, typically its something like SNAP_DT
                if not specified, will bypass checking step
        :return:
        """
        # do the checks to ensure successful concatenation
        li =  []
        columns = pd.Index([])
        for snap_dt in tqdm(snap_dts):
            df_partition = self.read(snap_dt=snap_dt, **kws)
            if partition_key is not None and partition_key not in df_partition.columns:
                raise ValueError(f"{self.schema}.{self.table}: Partition Key {partition_key} not found in {snap_dt}")
            if len(columns) > 0 and len(columns.union(df_partition.columns)) != len(columns):
                # if new column set has a different length than the previous one, means the columns are not same across all the table partitions
                raise ValueError(f"{self.schema}.{self.table}: Columns not consistent for snapshot {snap_dt}")
            li.append(df_partition)

        return pd.concat(li, axis = 0, ignore_index = True)
    
    def delete(self, snap_dt: date):
        """Delete a snap shot dataframe

        :param snap_dt: which snap date to delete
        :return:
        """
        filepath = self.get_default_file_path(snap_dt)
        self.sftp.delete(filepath)
        
    def drop(self):
        """drop the whole table

        :return:
        """
        self.sftp.delete_folder(self.dir)
        
    
    def duplicate(self, dst_schema: str, dst_table: str) -> SnapshotDataManagerSFTP:
        new = SnapshotDataManagerSFTP(schema = dst_schema, table = dst_table)
        for snap_dt in tqdm(self.get_existing_snap_dts()):
            from_filename = self.get_default_file_path(snap_dt=snap_dt)
            to_filename = new.get_default_file_path(snap_dt=snap_dt)
            self.sftp.copy(from_filename, to_filename)
        return new
    
    def disk_space(self, snap_dt, unit='MB') -> float:
        """get the size of the snap date file (pandas) or folder (pyspark partitions)

        :param snap_dt:
        :param unit: {KB, MB, GB}
        """
        f = self.get_default_file_path(snap_dt)
        size_bytes = self.sftp.size(f)

        scale = {'KB': 1, 'MB': 2, 'GB': 3}.get(unit, 0)
        size = size_bytes / (1024 ** scale)
        return size